# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

import numpy as np
import pytest

from stheno.eis import ComponentKernel, AdditiveComponentKernel, NoisyKernel
from stheno.input import Component, Latent, Observed
from stheno.kernel import EQ, RQ, ZeroKernel
from .util import allclose


def test_component_kernel():
    x = np.random.randn(10, 2)
    k1 = EQ()
    k2 = RQ(1e-1)
    kzero = ZeroKernel()
    kc = ComponentKernel({
        (Component(1), Component(1)): k1,
        (Component(1), Component(2)): kzero,
        (Component(2), Component(1)): kzero,
        (Component(2), Component(2)): k2
    })

    allclose(kc(Component(1)(x)), k1(x))
    allclose(kc(Component(2)(x)), k2(x))
    allclose(kc(Component(1)(x), Component(2)(x)), kzero(x))
    allclose(kc(Component(2)(x), Component(1)(x)), kzero(x))

    assert str(ComponentKernel({Component(1): EQ(),
                                Component(
                                    2): EQ()})) == 'ComponentKernel(EQ(), EQ())'


def test_compare_noisy_kernel_and_additive_component_kernel():
    def NoisyKernelCopy(k1, k2):
        return AdditiveComponentKernel({
            Component('wiggly'): k1,
            Component('noise'): k2
        }, latent=[Component('wiggly')])

    x = np.random.randn(10, 2)
    k1 = EQ()
    k2 = RQ(1e-2)
    kn1 = NoisyKernel(k1, k2)
    kn2 = NoisyKernelCopy(k1, k2)

    allclose(kn1(Latent(x)), k1(x), desc='noisy kernel 1')
    allclose(kn1(Latent(x), Observed(x)), k1(x), desc='noisy kernel 2')
    allclose(kn1(Observed(x), Latent(x)), k1(x), desc='noisy kernel 3')
    allclose(kn1(Observed(x)), (k1 + k2)(x), desc='noisy kernel 4')

    allclose(kn2(Latent(x)), k1(x), desc='noisy copy 1')
    allclose(kn2(Latent(x), Observed(x)), k1(x), desc='noisy copy 2')
    allclose(kn2(Observed(x), Latent(x)), k1(x), desc='noisy copy 3')
    allclose(kn2(Observed(x)), (k1 + k2)(x), desc='noisy copy 4')

    allclose(kn2(Latent(x), Component('noise')(x)), ZeroKernel()(x),
             desc='additive 1')
    allclose(kn2(Component('noise')(x), Latent(x)), ZeroKernel()(x),
             desc='additive 2')
    allclose(kn2(Observed(x), Component('noise')(x)), k2(x), desc='additive 3')
    allclose(kn2(Component('noise')(x), Observed(x)), k2(x), desc='additive 4')

    assert str(NoisyKernel(EQ(), RQ(1))) == 'NoisyKernel(EQ(), RQ(1))'


def test_additive_component_kernel():
    k = AdditiveComponentKernel({
        Component('wiggly'): EQ(),
        Component('noise'): EQ()
    })

    # Test independence components.
    x = np.random.randn(10, 2)
    allclose(k(Component('wiggly')(x), Component('noise')(x)),
             np.zeros((10, 10)))
    allclose(k(Component('noise')(x), Component('wiggly')(x)),
             np.zeros((10, 10)))

    # Test check.
    with pytest.raises(RuntimeError):
        k(Component('erroneous')(x))



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

import logging

from autograd.numpy.numpy_boxes import ArrayBox
from plum import Dispatcher

# noinspection PyUnresolvedReferences
from . import *  # pragma: no cover
from .field import Element
from .random import Random

__all__ = []

log = logging.getLogger(__name__)

_dispatch = Dispatcher()

# Save original methods.
__mul__array_box = ArrayBox.__mul__
__add__array_box = ArrayBox.__add__


@_dispatch(ArrayBox, object)
def __mul__(self, other):
    return __mul__array_box(self, other)


@_dispatch(ArrayBox, {Element, Random})
def __mul__(self, other):
    return other.__rmul__(self)


@_dispatch(ArrayBox, object)
def __add__(self, other):
    return __add__array_box(self, other)


@_dispatch(ArrayBox, {Element, Random})
def __add__(self, other):
    return other.__radd__(self)


# Assign dispatchable methods.
ArrayBox.__mul__ = __mul__
ArrayBox.__add__ = __add__



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

import logging

from plum import Dispatcher, Self, Referentiable

from .input import Observed, Latent, Component
from .kernel import Kernel, ZeroKernel

__all__ = ['NoisyKernel', 'ComponentKernel', 'AdditiveComponentKernel']

log = logging.getLogger(__name__)


class ComponentKernel(Kernel, Referentiable):
    """Kernel consisting of multiple components.

    Uses :class:`.input.Component`.

    Args:
        ks (dict[tuple[:class:`.input.Component`], :class:`.kernel.Kernel`]):
            Kernel between the various components, indexed by tuples of type
            parameters.
    """
    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, ks):
        self.ks = ks

    @_dispatch(Component, Component)
    def __call__(self, x, y):
        return self.ks[type(x), type(y)](x.get(), y.get())

    def __str__(self):
        ks = self.ks.values()
        return 'ComponentKernel({})'.format(', '.join([str(k) for k in ks]))


class AdditiveComponentKernel(ComponentKernel, Referentiable):
    """Kernel consisting of additive, independent components.

    Uses :class:`.input.Component`, :class:`.input.Observed`, and
    :class:`.input.Latent`.

    Args:
        ks (dict[:class:`.input.Component`, :class:`.kernel.Kernel`]): Kernels
            of the components, indexed by type parameters.
        latent (list[:class:`.input.Input`], optional): List of input types
            that make up the latent process.
    """

    def __init__(self, ks, latent=None):
        input_types = set(ks.keys())

        class KernelMatrix(object):
            def _resolve(self, t):
                if t == Latent:
                    return {} if latent is None else set(latent)
                elif t == Observed:
                    return input_types
                elif t in input_types:
                    return {t}
                else:
                    raise RuntimeError('Input type "{}" is not a component of '
                                       'the kernel.'.format(t.__name__))

            def __getitem__(self, item):
                t1s, t2s = self._resolve(item[0]), self._resolve(item[1])
                return sum([ks[t] for t in t1s & t2s], ZeroKernel())

        ComponentKernel.__init__(self, KernelMatrix())


class NoisyKernel(Kernel, Referentiable):
    """Noisy observations of a latent process.

    Uses :class:`.input.Latent` and :class:`.input.Observed`.

    Args:
        k_f (:class:`.kernel.Kernel`): Kernel of the latent process.
        k_n (:class:`.kernel.Kernel`): Kernel of the noise.
    """
    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, k_f, k_n):
        self.k_f = k_f
        self.k_n = k_n
        self.k_y = k_f + k_n

    @_dispatch({Latent, Observed}, {Latent, Observed})
    def __call__(self, x, y):
        return self.k_f(x.get(), y.get())

    @_dispatch(Observed, Observed)
    def __call__(self, x, y):
        return self.k_y(x.get(), y.get())

    def __str__(self):
        return 'NoisyKernel({}, {})'.format(self.k_f, self.k_n)



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

from plum import parametric, kind

__all__ = ['Input', 'Observed', 'Latent', 'Component', 'At', 'Unique']


class Input(object):
    """Input type.

    Args:
        *xs (tensor): Inputs to type.
    """

    def __init__(self, *xs):
        self._xs = xs

    def get(self):
        """Get the wrapped input."""
        return self._xs[0] if len(self._xs) == 1 else self._xs


@parametric
class Component(Input):
    """A particular component.

    This is a parametric type.
    """


class MultiInput(Input):
    """Multiple inputs."""

    def get(self):
        return self._xs


class Unique(Input):
    """One cannot learn about noise at these inputs"""


Observed = Component('observed')  #: Observed points
Latent = Component('latent')  #: Latent points

#: A generic parametric type used to specify locations at a particular process.
At = kind(Input)



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

import logging
from types import FunctionType

from fdm import central_fdm
from lab import B
from plum import Dispatcher, Self, Referentiable, type_parameter, Union

from .util import uprank
from .input import Input, At, MultiInput
from .kernel import ZeroKernel, PosteriorKernel, TensorProductKernel, \
    CorrectiveKernel, OneKernel
from .lazy import LazyVector, LazyMatrix
from .matrix import matrix, Diagonal, dense
from .mean import PosteriorMean, ZeroMean, OneMean
from .mokernel import MultiOutputKernel as MOK
from .momean import MultiOutputMean as MOM
from .random import Random, PromisedGP, RandomProcess, Normal

__all__ = ['GP',
           'model',
           'Graph',
           'AbstractObservations',
           'Observations', 'Obs',
           'SparseObservations', 'SparseObs']

log = logging.getLogger(__name__)


def ensure_at(x, ref=None):
    """Ensure that an input location is typed with `At` to specify which
    process it belongs to.

    Args:
        x (input): Input location.
        ref (:class:`.graph.GP`, optional): Reference process. If provided and
            `x` is not an instance of `At`, then it assumed to belong to `ref`.

    Returns:
        :class:`.input.At`: Input, instance of `At`.
    """
    if isinstance(x, At):
        return x
    elif ref is not None:
        return ref(x)
    else:
        raise ValueError('Must explicitly specify the processes which to '
                         'condition on.')


class AbstractObservations(Referentiable):
    """Abstract base class for observations."""

    _dispatch = Dispatcher(in_class=Self)

    @_dispatch({B.Numeric, Input}, B.Numeric, [PromisedGP])
    def __init__(self, x, y, ref=None):
        self._ref = ref
        self.x = ensure_at(x, self._ref)
        self.y = y
        self.graph = type_parameter(self.x).graph

    @_dispatch([Union(tuple, list, PromisedGP)])
    def __init__(self, *pairs, **kw_args):
        # Check whether there's a reference.
        self._ref = kw_args['ref'] if 'ref' in kw_args else None

        # Ensure `At` for all pairs.
        pairs = [(ensure_at(x, self._ref), y) for x, y in pairs]

        # Get the graph from the first pair.
        self.graph = type_parameter(pairs[0][0]).graph

        # Extend the graph by the Cartesian product `p` of all processes.
        p = self.graph.cross(*self.graph.ps)

        # Condition on the newly created vector-valued GP.
        xs, ys = zip(*pairs)
        self.x = p(MultiInput(*xs))
        self.y = B.concat(*[uprank(y) for y in ys], axis=0)

    @_dispatch({tuple, list})
    def __ror__(self, ps):
        return self.graph.condition(ps, self)

    def posterior_kernel(self, p_i, p_j):  # pragma: no cover
        """Get the posterior kernel between two processes.

        Args:
            p_i (:class:`.graph.GP`): First process.
            p_j (:class:`.graph.GP`): Second process.

        Returns:
            :class:`.kernel.Kernel`: Posterior kernel between the first and
                second process.
        """
        raise NotImplementedError('Posterior kernel construction not '
                                  'implemented.')

    def posterior_mean(self, p):  # pragma: no cover
        """Get the posterior kernel of a process.

        Args:
            p (:class:`.graph.GP`): Process.

        Returns:
            :class:`.mean.Mean`: Posterior mean of `p`.
        """
        raise NotImplementedError('Posterior mean construction not '
                                  'implemented.')


class Observations(AbstractObservations, Referentiable):
    """Observations.

    Can alternatively construct an instance of `Observations` with tuples or
    lists of valid constructors.

    Args:
        x (input): Locations of points to condition on.
        y (tensor): Observations to condition on.
        ref (:class:`.class.GP`, optional): Reference process. See
            :func:`.graph.ensure_at`.
    """
    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, *args, **kw_args):
        AbstractObservations.__init__(self, *args, **kw_args)
        self._K_x = None

    @property
    def K_x(self):
        """Kernel matrix of the data."""
        if self._K_x is None:  # Cache computation.
            p_x, x = type_parameter(self.x), self.x.get()
            self._K_x = matrix(self.graph.kernels[p_x](x))
        return self._K_x

    def posterior_kernel(self, p_i, p_j):
        p_x, x = type_parameter(self.x), self.x.get()
        return PosteriorKernel(self.graph.kernels[p_i, p_j],
                               self.graph.kernels[p_x, p_i],
                               self.graph.kernels[p_x, p_j],
                               x, self.K_x)

    def posterior_mean(self, p):
        p_x, x = type_parameter(self.x), self.x.get()
        return PosteriorMean(self.graph.means[p],
                             self.graph.means[p_x],
                             self.graph.kernels[p_x, p],
                             x, self.K_x, self.y)


class SparseObservations(AbstractObservations, Referentiable):
    """Observations through inducing points. Takes further arguments
    according to the constructor of :class:`.graph.Observations`.

    Attributes:
        elbo (scalar): ELBO.

    Args:
        z (input): Locations of the inducing points.
        e (:class:`.graph.GP`): Additive, independent noise process.
    """

    _dispatch = Dispatcher(in_class=Self)

    @_dispatch({B.Numeric, Input, tuple, list},
               [Union(tuple, list, PromisedGP)])
    def __init__(self, z, *pairs, **kw_args):
        es, xs, ys = zip(*pairs)
        AbstractObservations.__init__(self, *zip(xs, ys), **kw_args)
        SparseObservations.__init__(self,
                                    z,
                                    self.graph.cross(*es),
                                    self.x,
                                    self.y,
                                    **kw_args)

    @_dispatch({list, tuple},
               PromisedGP,
               {B.Numeric, Input},
               B.Numeric,
               [PromisedGP])
    def __init__(self, zs, e, x, y, ref=None):
        # Ensure `At` everywhere.
        zs = [ensure_at(z, ref=ref) for z in zs]

        # Extract graph.
        graph = type_parameter(zs[0]).graph

        # Create a representative multi-output process.
        p_z = graph.cross(*(type_parameter(z) for z in zs))

        SparseObservations.__init__(self,
                                    p_z(MultiInput(*zs)),
                                    e, x, y, ref=ref)

    @_dispatch({B.Numeric, Input},
               PromisedGP,
               {B.Numeric, Input},
               B.Numeric,
               [PromisedGP])
    def __init__(self, z, e, x, y, ref=None):
        AbstractObservations.__init__(self, x, y, ref=ref)
        self.z = ensure_at(z, self._ref)
        self.e = e

        self._K_z = None
        self._elbo = None
        self._mu = None
        self._A = None

    @property
    def K_z(self):
        """Kernel matrix of the data."""
        if self._K_z is None:  # Cache computation.
            self._compute()
        return self._K_z

    @property
    def elbo(self):
        """ELBO."""
        if self._elbo is None:  # Cache computation.
            self._compute()
        return self._elbo

    @property
    def mu(self):
        """Mean of optimal approximating distribution."""
        if self._mu is None:  # Cache computation.
            self._compute()
        return self._mu

    @property
    def A(self):
        """Parameter of the corrective variance of the kernel of the optimal
        approximating distribution."""
        if self._A is None:  # Cache computation.
            self._compute()
        return self._A

    def _compute(self):
        # Extract processes.
        p_x, x = type_parameter(self.x), self.x.get()
        p_z, z = type_parameter(self.z), self.z.get()

        # Construct the necessary kernel matrices.
        K_zx = self.graph.kernels[p_z, p_x](z, x)
        self._K_z = matrix(self.graph.kernels[p_z](z))

        # Evaluating `e.kernel(x)` will yield incorrect results if `x` is a
        # `MultiInput`, because `x` then still designates the particular
        # components of `f`. Fix that by instead designating the elements of
        # `e`.
        if isinstance(x, MultiInput):
            x_n = MultiInput(*(p(xi.get())
                               for p, xi in zip(self.e.kernel.ps, x.get())))
        else:
            x_n = x

        # Construct the noise kernel matrix.
        K_n = self.e.kernel(x_n)

        # The approximation can only handle diagonal noise matrices.
        if not isinstance(K_n, Diagonal):
            raise RuntimeError('Kernel matrix of noise must be diagonal.')

        # And construct the components for the inducing point approximation.
        L_z = B.cholesky(self._K_z)
        self._A = B.eye(self._K_z) + \
                  B.qf(K_n, B.transpose(B.trisolve(L_z, K_zx)))
        y_bar = uprank(self.y) - self.e.mean(x_n) - self.graph.means[p_x](x)
        prod_y_bar = B.trisolve(L_z, B.qf(K_n, B.transpose(K_zx), y_bar))

        # Compute the optimal mean.
        self._mu = self.graph.means[p_z](z) + \
                   B.qf(self._A, B.trisolve(L_z, self._K_z), prod_y_bar)

        # Compute the ELBO.
        # NOTE: The calculation of `trace_part` asserts that `K_n` is diagonal.
        #       The rest, however, is completely generic.
        trace_part = B.ratio(Diagonal(self.graph.kernels[p_x].elwise(x)[:, 0]) -
                             Diagonal(B.qf_diag(self._K_z, K_zx)), K_n)
        det_part = B.logdet(2 * B.pi * K_n) + B.logdet(self._A)
        qf_part = B.qf(K_n, y_bar)[0, 0] - B.qf(self._A, prod_y_bar)[0, 0]
        self._elbo = -0.5 * (trace_part + det_part + qf_part)

    def posterior_kernel(self, p_i, p_j):
        p_z, z = type_parameter(self.z), self.z.get()
        return PosteriorKernel(self.graph.kernels[p_i, p_j],
                               self.graph.kernels[p_z, p_i],
                               self.graph.kernels[p_z, p_j],
                               z, self.K_z) + \
               CorrectiveKernel(self.graph.kernels[p_z, p_i],
                                self.graph.kernels[p_z, p_j],
                                z, self.A, self.K_z)

    def posterior_mean(self, p):
        p_z, z = type_parameter(self.z), self.z.get()
        return PosteriorMean(self.graph.means[p],
                             self.graph.means[p_z],
                             self.graph.kernels[p_z, p],
                             z, self.K_z, self.mu)


Obs = Observations  #: Shorthand for `Observations`.
SparseObs = SparseObservations  #: Shorthand for `SparseObservations`.


class Graph(Referentiable):
    """A GP model."""
    _dispatch = Dispatcher(in_class=Self)

    def __init__(self):
        self.ps = []
        self.pids = set()
        self.kernels = LazyMatrix()
        self.means = LazyVector()

        # Store named GPs in both ways.
        self.gps_by_name = {}
        self.names_by_gp = {}

    @_dispatch(str)
    def __getitem__(self, name):
        return self.gps_by_name[name]

    @_dispatch(PromisedGP)
    def __getitem__(self, p):
        return self.names_by_gp[id(p)]

    @_dispatch(PromisedGP, str)
    def name(self, p, name):
        """Name a GP.

        Args:
            p (:class:`.graph.GP`): GP to name.
            name (str): Name. Must be unique.
        """
        # Delete any existing names and back-references for the GP.
        if id(p) in self.names_by_gp:
            del self.gps_by_name[self.names_by_gp[id(p)]]
            del self.names_by_gp[id(p)]

        # Check that name is not in use.
        if name in self.gps_by_name:
            raise RuntimeError('Name "{}" for "{}" already taken by "{}".'
                               ''.format(name, p, self[name]))

        # Set the name and the back-reference.
        self.gps_by_name[name] = p
        self.names_by_gp[id(p)] = name

    def _add_p(self, p):
        self.ps.append(p)
        self.pids.add(id(p))

    def _update(self, mean, k_ii_generator, k_ij_generator):
        p = GP(self)
        self.means[p] = mean
        self.kernels.add_rule((p, p), self.pids, k_ii_generator)
        self.kernels.add_rule((p, None), self.pids, k_ij_generator)
        self.kernels.add_rule((None, p), self.pids,
                              lambda pi: reversed(self.kernels[p, pi]))
        self._add_p(p)
        return p

    def add_independent_gp(self, p, kernel, mean):
        """Add an independent GP to the model.

        Args:
            p (:class:`.graph.GP`): GP object to add.
            kernel (:class:`.kernel.Kernel`): Kernel function of GP.
            mean (:class:`.mean.Mean`): Mean function of GP.

        Returns:
            :class:`.graph.GP`: The newly added independent GP.
        """
        # Update means.
        self.means[p] = mean
        # Add rule to kernels.
        self.kernels[p] = kernel
        self.kernels.add_rule((p, None), self.pids, lambda pi: ZeroKernel())
        self.kernels.add_rule((None, p), self.pids, lambda pi: ZeroKernel())
        self._add_p(p)
        return p

    @_dispatch(object, PromisedGP)
    def sum(self, other, p):
        """Sum a GP from the graph with another object.

        Args:
            obj1 (other type or :class:`.graph.GP`): First term in the sum.
            obj2 (other type or :class:`.graph.GP`): Second term in the sum.

        Returns:
            :class:`.graph.GP`: The GP corresponding to the sum.
        """
        return self.sum(p, other)

    @_dispatch(PromisedGP, object)
    def sum(self, p, other):
        return self._update(self.means[p] + other,
                            lambda: self.kernels[p],
                            lambda pi: self.kernels[p, pi])

    @_dispatch(PromisedGP, PromisedGP)
    def sum(self, p1, p2):
        # Check that the GPs are on the same graph.
        if p1.graph != p2.graph:
            raise RuntimeError('Can only add GPs from the same graph.')

        return self._update(self.means[p1] + self.means[p2],
                            (lambda: self.kernels[p1] +
                                     self.kernels[p2] +
                                     self.kernels[p1, p2] +
                                     self.kernels[p2, p1]),
                            lambda pi: self.kernels[p1, pi] +
                                       self.kernels[p2, pi])

    @_dispatch(PromisedGP, B.Numeric)
    def mul(self, p, other):
        """Multiply a GP from the graph with another object.

        Args:
            p (:class:`.graph.GP`): GP in the product.
            other (object): Other object in the product.

        Returns:
            :class:`.graph.GP`: The GP corresponding to the product.
        """
        return self._update(self.means[p] * other,
                            lambda: self.kernels[p] * other ** 2,
                            lambda pi: self.kernels[p, pi] * other)

    @_dispatch(PromisedGP, FunctionType)
    def mul(self, p, f):
        def ones(x):
            return B.ones(B.dtype(x), B.shape(x)[0], 1)

        return self._update(f * self.means[p],
                            lambda: f * self.kernels[p],
                            (lambda pi: TensorProductKernel(f, ones) *
                                        self.kernels[p, pi]))

    def shift(self, p, shift):
        """Shift a GP.

        Args:
            p (:class:`.graph.GP`): GP to shift.
            shift (object): Amount to shift by.

        Returns:
            :class:`.graph.GP`: The shifted GP.
        """
        return self._update(self.means[p].shift(shift),
                            lambda: self.kernels[p].shift(shift),
                            lambda pi: self.kernels[p, pi].shift(shift, 0))

    def stretch(self, p, stretch):
        """Stretch a GP.

        Args:
            p (:class:`.graph.GP`): GP to stretch.
            stretch (object): Extent of stretch.

        Returns:
            :class:`.graph.GP`: The stretched GP.
        """
        return self._update(self.means[p].stretch(stretch),
                            lambda: self.kernels[p].stretch(stretch),
                            lambda pi: self.kernels[p, pi].stretch(stretch, 1))

    def select(self, p, *dims):
        """Select input dimensions.

        Args:
            p (:class:`.graph.GP`): GP to select input
                dimensions from.
            *dims (object): Dimensions to select.

        Returns:
            :class:`.graph.GP`: GP with the specific input dimensions.
        """
        return self._update(self.means[p].select(dims),
                            lambda: self.kernels[p].select(dims),
                            lambda pi: self.kernels[p, pi].select(dims, None))

    def transform(self, p, f):
        """Transform the inputs of a GP.

        Args:
            p (:class:`.graph.GP`): GP to input transform.
            f (function): Input transform.

        Returns:
            :class:`.graph.GP`: Input-transformed GP.
        """
        return self._update(self.means[p].transform(f),
                            lambda: self.kernels[p].transform(f),
                            lambda pi: self.kernels[p, pi].transform(f, None))

    def diff(self, p, dim=0):
        """Differentiate a GP.

        Args:
            p (:class:`.graph.GP`): GP to differentiate.
            dim (int, optional): Dimension of feature which to take the
                derivative with respect to. Defaults to `0`.

        Returns:
            :class:`.graph.GP`: Derivative of GP.
        """
        return self._update(self.means[p].diff(dim),
                            lambda: self.kernels[p].diff(dim),
                            lambda pi: self.kernels[p, pi].diff(dim, None))

    @_dispatch({list, tuple}, AbstractObservations)
    def condition(self, ps, obs):
        """Condition the graph on observations.

        Args:
            ps (list[:class:`.graph.GP`]): Processes to condition.
            obs (:class:`.graph.AbstractObservations`): Observations to
                condition on.

        Returns:
            list[:class:`.graph.GP`]: Posterior processes.
        """

        # A construction like this is necessary to properly close over `p`.
        def build_gens(p):
            def k_ij_generator(pi):
                return obs.posterior_kernel(p, pi)

            def k_ii_generator():
                return obs.posterior_kernel(p, p)

            return k_ii_generator, k_ij_generator

        return [self._update(obs.posterior_mean(p), *build_gens(p)) for p in ps]

    def cross(self, *ps):
        """Construct the Cartesian product of a collection of processes.

        Args:
            *ps (:class:`.graph.GP`): Processes to construct the
                Cartesian product of.

        Returns:
            :class:`.graph.GP`: The Cartesian product of `ps`.
        """
        mok = MOK(*ps)
        return self._update(MOM(*ps),
                            lambda: mok,
                            lambda pi: mok.transform(None, lambda y: At(pi)(y)))

    @_dispatch(int, [At])
    def sample(self, n, *xs):
        """Sample multiple processes simultaneously.

        Args:
            n (int, optional): Number of samples. Defaults to `1`.
            *xs (:class:`.graph.At`): Locations to sample at.

        Returns:
            tuple: Tuple of samples.
        """
        sample = GP(MOK(*self.ps),
                    MOM(*self.ps),
                    graph=Graph())(MultiInput(*xs)).sample(n)

        # To unpack `x`, just keep `.get()`ing.
        def unpack(x):
            while isinstance(x, Input):
                x = x.get()
            return x

        # Unpack sample.
        lengths = [B.shape(uprank(unpack(x)))[0] for x in xs]
        i, samples = 0, []
        for length in lengths:
            samples.append(sample[i:i + length, :])
            i += length
        return samples[0] if len(samples) == 1 else samples

    @_dispatch([At])
    def sample(self, *xs):
        return self.sample(1, *xs)

    @_dispatch([{list, tuple}])
    def logpdf(self, *pairs):
        xs, ys = zip(*pairs)

        # Check that all processes are specified.
        if not all([isinstance(x, At) for x in xs]):
            raise ValueError('Must explicitly specify the processes which to '
                             'compute the log-pdf for.')

        # Uprank all outputs and concatenate.
        y = B.concat(*[uprank(y) for y in ys], axis=0)

        # Return composite log-pdf.
        return GP(MOK(*self.ps),
                  MOM(*self.ps),
                  graph=Graph())(MultiInput(*xs)).logpdf(y)

    @_dispatch(At, B.Numeric)
    def logpdf(self, x, y):
        return x.logpdf(y)

    @_dispatch(Observations)
    def logpdf(self, obs):
        return obs.x.logpdf(obs.y)

    @_dispatch(SparseObservations)
    def logpdf(self, obs):
        return obs.elbo


model = Graph()  #: A default graph provided for convenience


class GP(RandomProcess, Referentiable):
    """Gaussian process.

    Args:
        kernel (:class:`.kernel.Kernel`): Kernel of the
            process.
        mean (:class:`.mean.Mean`, optional): Mean function of the
            process. Defaults to zero.
        graph (:class:`.graph.Graph`, optional): Graph to attach to.
    """
    _dispatch = Dispatcher(in_class=Self)

    @_dispatch([object])
    def __init__(self, kernel, mean=None, graph=model, name=None):
        # Resolve kernel.
        if isinstance(kernel, (B.Numeric, FunctionType)):
            kernel = kernel * OneKernel()

        # Resolve mean.
        if mean is None:
            mean = ZeroMean()
        elif isinstance(mean, (B.Numeric, FunctionType)):
            mean = mean * OneMean()

        # Then add a new `GP` to the graph with the resolved kernel and mean.
        self.graph = graph
        self.graph.add_independent_gp(self, kernel, mean)

        # If a name is given, set the name.
        if name:
            self.graph.name(self, name)

    @_dispatch(Graph)
    def __init__(self, graph):
        self.graph = graph

    @property
    def kernel(self):
        """Kernel of the GP."""
        return self.graph.kernels[self]

    @property
    def mean(self):
        """Mean function of the GP."""
        return self.graph.means[self]

    @property
    def name(self):
        """Name of the GP."""
        return self.graph[self]

    @name.setter
    @_dispatch(str)
    def name(self, name):
        self.graph.name(self, name)

    def __call__(self, x):
        """Construct a finite-dimensional distribution at specified locations.

        Args:
            x (input): Points to construct the distribution at.

        Returns:
            :class:`.random.Normal`: Finite-dimensional distribution.
        """
        return Normal(self, x)

    @_dispatch([object])
    def condition(self, *args):
        """Condition the GP. See :meth:`.graph.Graph.condition`."""
        return self.graph.condition((self,), Observations(*args, ref=self))[0]

    @_dispatch(AbstractObservations)
    def condition(self, obs):
        return self.graph.condition((self,), obs)[0]

    @_dispatch(object)
    def __add__(self, other):
        return self.graph.sum(self, other)

    @_dispatch(Random)
    def __add__(self, other):
        raise NotImplementedError('Cannot add a GP and a {}.'
                                  ''.format(type(other).__name__))

    @_dispatch(Self)
    def __add__(self, other):
        return self.graph.sum(self, other)

    @_dispatch(object)
    def __mul__(self, other):
        return self.graph.mul(self, other)

    @_dispatch(Random)
    def __mul__(self, other):
        raise NotImplementedError('Cannot multiply a GP and a {}.'
                                  ''.format(type(other).__name__))

    @_dispatch(Self)
    def __mul__(self, other):
        return (lambda x: self.graph.means[self](x)) * other + \
               self * (lambda x: self.graph.means[other](x)) + \
               GP(kernel=self.graph.kernels[self] *
                         self.graph.kernels[other] +
                         self.graph.kernels[self, other] *
                         self.graph.kernels[other, self],
                  mean=-self.graph.means[self] *
                       self.graph.means[other],
                  graph=self.graph)

    @_dispatch([object])
    def __or__(self, args):
        """Shorthand for conditioning."""
        return self.condition(Observations(*args, ref=self))

    @_dispatch(AbstractObservations)
    def __or__(self, obs):
        return self.condition(obs)

    def shift(self, shift):
        """Shift the GP. See :meth:`.graph.Graph.shift`."""
        return self.graph.shift(self, shift)

    def stretch(self, stretch):
        """Stretch the GP. See :meth:`.graph.Graph.stretch`."""
        return self.graph.stretch(self, stretch)

    def __gt__(self, stretch):
        """Shorthand for :meth:`.graph.GP.stretch`."""
        return self.stretch(stretch)

    def transform(self, f):
        """Input transform the GP. See :meth:`.graph.Graph.transform`."""
        return self.graph.transform(self, f)

    def select(self, *dims):
        """Select dimensions from the input. See :meth:`.graph.Graph.select`."""
        return self.graph.select(self, *dims)

    def __getitem__(self, *dims):
        """Shorthand for :meth:`.graph.GP.select`."""
        return self.select(*dims)

    def diff(self, dim=0):
        """Differentiate the GP. See :meth:`.graph.Graph.diff`."""
        return self.graph.diff(self, dim)

    def diff_approx(self, deriv=1, order=6):
        """Approximate the derivative of the GP by constructing a finite
        difference approximation.

        Args:
            deriv (int): Order of the derivative.
            order (int): Order of the estimate.

        Returns:
            Approximation of the derivative of the GP.
        """
        # Use the FDM library to figure out the coefficients.
        fdm = central_fdm(order, deriv, adapt=0, factor=1e8)
        fdm.estimate()  # Estimate step size.

        # Construct finite difference.
        df = 0
        for g, c in zip(fdm.grid, fdm.coefs):
            df += c * self.shift(-g * fdm.step)
        return df / fdm.step ** deriv

    @property
    def stationary(self):
        """Stationarity of the GP."""
        return self.kernel.stationary

    @property
    def var(self):
        """Variance of the GP."""
        return self.kernel.var

    @property
    def length_scale(self):
        """Length scale of the GP."""
        return self.kernel.length_scale

    @property
    def period(self):
        """Period of the GP."""
        return self.kernel.period

    def __str__(self):
        return self.display()

    def __repr__(self):
        return self.display()

    def display(self, formatter=lambda x: x):
        """Display the GP.

        Args:
            formatter (function, optional): Function to format values.

        Returns:
            str: GP as a string.
        """
        return 'GP({}, {})'.format(self.kernel.display(formatter),
                                   self.mean.display(formatter))


PromisedGP.deliver(GP)



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

from numbers import Number

import numpy as np
import pytest
import tensorflow as tf
from plum import Dispatcher
import lab.tensorflow as B

from stheno.input import Observed
from stheno.kernel import EQ
from stheno.matrix import matrix
from stheno.mean import (
    TensorProductMean,
    ZeroMean,
    Mean,
    OneMean,
    PosteriorMean
)
from .util import allclose


def test_corner_cases():
    with pytest.raises(NotImplementedError):
        Mean()(1.)


def test_construction():
    m = TensorProductMean(lambda x: x ** 2)
    x = np.random.randn(10, 1)

    m(x)
    m(Observed(x))


def test_basic_arithmetic():
    dispatch = Dispatcher()

    @dispatch(Number)
    def f1(x): return np.array([[x ** 2]])

    @dispatch(object)
    def f1(x): return np.sum(x ** 2, axis=1)[:, None]

    @dispatch(Number)
    def f2(x): return np.array([[x ** 3]])

    @dispatch(object)
    def f2(x): return np.sum(x ** 3, axis=1)[:, None]

    m1 = TensorProductMean(f1)
    m2 = TensorProductMean(f2)
    m3 = ZeroMean()
    x1 = np.random.randn(10, 2)
    x2 = np.random.randn()

    assert np.allclose((m1 * m2)(x1), m1(x1) * m2(x1)), 'prod'
    assert np.allclose((m1 * m2)(x2), m1(x2) * m2(x2)), 'prod 2'
    assert np.allclose((m1 + m3)(x1), m1(x1) + m3(x1)), 'sum'
    assert np.allclose((m1 + m3)(x2), m1(x2) + m3(x2)), 'sum 2'
    assert np.allclose((5. * m1)(x1), 5. * m1(x1)), 'prod 3'
    assert np.allclose((5. * m1)(x2), 5. * m1(x2)), 'prod 4'
    assert np.allclose((5. + m1)(x1), 5. + m1(x1)), 'sum 3'
    assert np.allclose((5. + m1)(x2), 5. + m1(x2)), 'sum 4'


def test_posterior_mean():
    z = np.linspace(0, 1, 10)
    pcm = PosteriorMean(
        TensorProductMean(lambda x: x),
        TensorProductMean(lambda x: x ** 2),
        EQ(), z, matrix(2 * EQ()(z)), np.random.randn(10)
    )

    # Check name.
    assert str(pcm) == 'PosteriorMean()'

    # Check that the mean computes.
    pcm(z)


def test_function_mean():
    m1 = 5 * OneMean() + (lambda x: x ** 2)
    m2 = (lambda x: x ** 2) + 5 * OneMean()
    m3 = (lambda x: x ** 2) + ZeroMean()
    m4 = ZeroMean() + (lambda x: x ** 2)
    x = np.random.randn(10, 1)

    assert np.allclose(m1(x), 5 + x ** 2)
    assert np.allclose(m2(x), 5 + x ** 2)
    assert np.allclose(m3(x), x ** 2)
    assert np.allclose(m4(x), x ** 2)

    def my_function(x): pass

    assert str(TensorProductMean(my_function)) == 'my_function'


def test_derivative():
    m = TensorProductMean(lambda x: x ** 2)
    m2 = TensorProductMean(lambda x: x ** 3)
    x = B.randn(tf.float64, 10, 1)

    allclose(m.diff(0)(x), 2 * x)
    allclose(m2.diff(0)(x), 3 * x ** 2)


def test_selected_mean():
    m = 5 * OneMean() + (lambda x: x ** 2)
    x = np.random.randn(10, 3)

    allclose(m.select([1, 2])(x), m(x[:, [1, 2]]))


def test_shifting():
    m = 5 * OneMean() + (lambda x: x ** 2)
    x = np.random.randn(10, 3)

    allclose(m.shift(5)(x), m(x - 5))


def test_stretching():
    m = 5 * OneMean() + (lambda x: x ** 2)
    x = np.random.randn(10, 3)

    allclose(m.stretch(5)(x), m(x / 5))


def test_input_transform():
    m = 5 * OneMean() + (lambda x: x ** 2)
    x = np.random.randn(10, 3)

    allclose(m.transform(lambda x: x - 5)(x), m(x - 5))



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

import pytest
import lab as B
import tensorflow as tf
import torch
from autograd import grad
from lab import torch as B

from stheno.field import Element
from stheno.torch import EQ


def test_boxing_autograd():
    k = EQ()
    objs = []

    def objective(x):
        x = x + 2
        x = x * 2
        objs.append(x + k)
        objs.append(x + k(x))
        objs.append(x * k)
        objs.append(x * k(x))
        return B.sum(x)

    grad(objective)(B.randn(10))

    for obj in objs:
        assert isinstance(obj, Element)


@pytest.mark.parametrize('dtype, Numeric',
                         [(torch.float64, B.TorchNumeric),
                          (tf.float64, B.TFNumeric)])
def test_boxing(dtype, Numeric):
    k = EQ()
    num = B.randn(dtype)
    x = B.randn(dtype, 10)

    # Test addition and multiplication.
    assert isinstance(num + k, Element)
    assert isinstance(num + k(x), Element)
    assert isinstance(num + num, Numeric)
    assert isinstance(num * k, Element)
    assert isinstance(num * k(x), Element)
    assert isinstance(num * num, Numeric)

    # Test in-place addition.
    num = B.randn(dtype)
    num += B.randn(dtype)
    assert isinstance(num, Numeric)

    num += k
    assert isinstance(num, Element)

    num = B.randn(dtype)
    num += k(x)
    assert isinstance(num, Element)

    # Test in-place multiplication.
    num = B.randn(dtype)
    num *= B.randn(dtype)
    assert isinstance(num, Numeric)

    num *= k
    assert isinstance(num, Element)

    num = B.randn(dtype)
    num *= k(x)
    assert isinstance(num, Element)



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

import pytest

from stheno.lazy import Rule, LazyMatrix, LazyVector


def test_corner_cases():
    assert repr(Rule(1, {1}, 1)) == \
           'Rule(pattern=1, indices={!r}, builder=1)'.format({1})


def test_indexing():
    v = LazyVector()
    m = LazyMatrix()

    v[1] = 1
    assert v[1] == 1

    # Test diagonal.
    m[1] = 1
    m[2, 2] = 2
    assert m[1, 1] == 1
    assert m[2] == 2

    # Test resolving of indices.
    class A(object):
        pass

    a1, a2 = A(), A()
    m[id(a1), a2] = (1, 2)
    m[id(a1)] = 1
    m[a2] = 2

    assert m[a1, a1] == 1
    assert m[a1, a2] == (1, 2)
    assert m[a2, a2] == 2


def test_rule():
    rule = Rule((1, None, 2, None, 3), {1, 2, 3, 4, 5}, lambda: None)

    assert rule.applies((1, 1, 2, 5, 3))
    assert not rule.applies((1, 1, 2, 6, 3))
    assert not rule.applies((2, 1, 2, 5, 3))


def test_building():
    class ReversibleNumber(object):
        def __init__(self, x):
            self.x = x

        def __reversed__(self):
            return ReversibleNumber(self.x)

    m = LazyMatrix()

    # Test universal building.
    m.add_rule((None, None), range(3), lambda x, y: ReversibleNumber(x * y))

    for i in range(3):
        for j in range(3):
            assert m[i, j].x == i * j

    with pytest.raises(RuntimeError):
        m[-1, 0]
    with pytest.raises(RuntimeError):
        m[0, -1]
    with pytest.raises(RuntimeError):
        m[-1, -1]
    with pytest.raises(RuntimeError):
        m[4, 3]
    with pytest.raises(RuntimeError):
        m[3, 4]
    with pytest.raises(RuntimeError):
        m[4, 4]

    # Test building along first dimension.
    m.add_rule((3, None), range(4), lambda y: ReversibleNumber(3 + y))
    m.add_rule((None, 3), range(4), lambda y: ReversibleNumber(3 + y))

    for i in range(3):
        for j in range(3):
            assert m[i, j].x == i * j

    for i in range(3):
        assert m[i, 3].x == i + 3
        assert m[3, i].x == 3 + i
    assert m[3, 3].x == 3 + 3

    with pytest.raises(RuntimeError):
        m[-1, 0]
    with pytest.raises(RuntimeError):
        m[0, -1]
    with pytest.raises(RuntimeError):
        m[-1, -1]
    with pytest.raises(RuntimeError):
        m[5, 4]
    with pytest.raises(RuntimeError):
        m[4, 5]
    with pytest.raises(RuntimeError):
        m[5, 5]

    # Test building along second dimension.
    m.add_rule((None, 4), range(5), lambda x: ReversibleNumber(x ** 2 + 4 ** 2))
    m.add_rule((4, None), range(5), lambda x: ReversibleNumber(x ** 2 + 4 ** 2))

    for i in range(3):
        for j in range(3):
            assert m[i, j].x == i * j

    for i in range(3):
        assert m[i, 3].x == i + 3
        assert m[3, i].x == 3 + i
    assert m[3, 3].x == 3 + 3

    for i in range(4):
        assert m[i, 4].x == i ** 2 + 4 ** 2
        assert m[4, i].x == 4 ** 2 + i ** 2
    assert m[4, 4].x == 4 ** 2 + 4 ** 2

    with pytest.raises(RuntimeError):
        m[-1, 0]
    with pytest.raises(RuntimeError):
        m[0, -1]
    with pytest.raises(RuntimeError):
        m[-1, -1]
    with pytest.raises(RuntimeError):
        m[6, 5]
    with pytest.raises(RuntimeError):
        m[5, 6]
    with pytest.raises(RuntimeError):
        m[6, 6]

    # Now try all rules at once and mix up access order.
    m2 = LazyMatrix()
    m2.add_rule((None, None), range(3), lambda x, y: ReversibleNumber(x * y))
    m2.add_rule((3, None), range(4), lambda y: ReversibleNumber(3 + y))
    m2.add_rule((None, 3), range(4), lambda y: ReversibleNumber(3 + y))
    m2.add_rule((None, 4), range(5),
                lambda x: ReversibleNumber(x ** 2 + 4 ** 2))
    m2.add_rule((4, None), range(5),
                lambda x: ReversibleNumber(x ** 2 + 4 ** 2))

    for i in range(4):
        assert m2[i, 4].x == i ** 2 + 4 ** 2
        assert m2[4, i].x == 4 ** 2 + i ** 2
    assert m2[4, 4].x == 4 ** 2 + 4 ** 2

    for i in range(3):
        assert m2[i, 3].x == i + 3
        assert m2[3, i].x == 3 + i
    assert m2[3, 3].x == 3 + 3

    for i in range(3):
        for j in range(3):
            assert m2[i, j].x == i * j

    with pytest.raises(RuntimeError):
        m2[-1, 0]
    with pytest.raises(RuntimeError):
        m2[0, -1]
    with pytest.raises(RuntimeError):
        m2[-1, -1]
    with pytest.raises(RuntimeError):
        m2[6, 5]
    with pytest.raises(RuntimeError):
        m2[5, 6]
    with pytest.raises(RuntimeError):
        m2[6, 6]



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

import numpy as np
import pytest
import tensorflow as tf
from lab import B
from plum import type_parameter

from stheno.graph import Graph, GP, Obs, SparseObs
from stheno.input import At, Unique
from stheno.kernel import (
    Linear,
    EQ,
    Delta,
    Exp,
    RQ,
    ZeroKernel,
    OneKernel,
    ScaledKernel
)
from stheno.mean import TensorProductMean, ZeroMean, ScaledMean, OneMean
from stheno.random import Normal
from .util import allclose


def abs_err(x1, x2=0):
    return np.sum(np.abs(x1 - x2))


def rel_err(x1, x2):
    return 2 * abs_err(x1, x2) / (abs_err(x1) + abs_err(x2))


def test_corner_cases():
    m1 = Graph()
    m2 = Graph()
    p1 = GP(EQ(), graph=m1)
    p2 = GP(EQ(), graph=m2)
    x = np.random.randn(10, 2)
    with pytest.raises(RuntimeError):
        p1 + p2
    with pytest.raises(NotImplementedError):
        p1 + p1(x)
    with pytest.raises(NotImplementedError):
        p1 * p1(x)
    assert str(GP(EQ(), graph=m1)) == 'GP(EQ(), 0)'


def test_construction():
    model = Graph()
    p = GP(EQ(), graph=model)

    x = np.random.randn(10, 1)

    p.mean(x)
    p.mean(p(x))

    p.kernel(x)
    p.kernel(p(x))
    p.kernel(x, x)
    p.kernel(p(x), x)
    p.kernel(x, p(x))
    p.kernel(p(x), p(x))

    p.kernel.elwise(x)
    p.kernel.elwise(p(x))
    p.kernel.elwise(x, x)
    p.kernel.elwise(p(x), x)
    p.kernel.elwise(x, p(x))
    p.kernel.elwise(p(x), p(x))

    # Test resolution of kernel and mean.
    k = EQ()
    m = TensorProductMean(lambda x: x ** 2)

    assert isinstance(GP(k, graph=model).mean, ZeroMean)
    assert isinstance(GP(k, 5, graph=model).mean, ScaledMean)
    assert isinstance(GP(k, 1, graph=model).mean, OneMean)
    assert isinstance(GP(k, 0, graph=model).mean, ZeroMean)
    assert isinstance(GP(k, m, graph=model).mean, TensorProductMean)
    assert isinstance(GP(k, graph=model).kernel, EQ)
    assert isinstance(GP(5, graph=model).kernel, ScaledKernel)
    assert isinstance(GP(1, graph=model).kernel, OneKernel)
    assert isinstance(GP(0, graph=model).kernel, ZeroKernel)

    # Test construction of finite-dimensional distribution.
    d = GP(k, m, graph=model)(x)
    allclose(d.var, k(x))
    allclose(d.mean, m(x))


def test_sum_other():
    model = Graph()
    p1 = GP(EQ(), TensorProductMean(lambda x: x ** 2), graph=model)
    p2 = p1 + 5.
    p3 = 5. + p1
    p4 = model.sum(5., p1)

    x = np.random.randn(5, 1)
    allclose(p1.mean(x) + 5., p2.mean(x))
    allclose(p1.mean(x) + 5., p3.mean(x))
    allclose(p1.mean(x) + 5., p4.mean(x))
    allclose(p1.kernel(x), p2.kernel(x))
    allclose(p1.kernel(x), p3.kernel(x))
    allclose(p1.kernel(x), p4.kernel(x))
    allclose(p1.kernel(p2(x), p3(x)), p1.kernel(x))
    allclose(p1.kernel(p2(x), p4(x)), p1.kernel(x))

    # Check that a `GP` cannot be summed with a `Normal`.
    with pytest.raises(NotImplementedError):
        p1 + Normal(np.eye(3))
    with pytest.raises(NotImplementedError):
        Normal(np.eye(3)) + p1


def test_mul_other():
    model = Graph()
    p1 = GP(EQ(), TensorProductMean(lambda x: x ** 2), graph=model)
    p2 = 5. * p1
    p3 = p1 * 5.

    x = np.random.randn(5, 1)
    allclose(5. * p1.mean(x), p2.mean(x))
    allclose(5. * p1.mean(x), p3.mean(x))
    allclose(25. * p1.kernel(x), p2.kernel(x))
    allclose(25. * p1.kernel(x), p3.kernel(x))
    allclose(model.kernels[p2, p3](x, x), 25. * p1.kernel(x))

    # Check that a `GP` cannot be multiplied with a `Normal`.
    with pytest.raises(NotImplementedError):
        p1 * Normal(np.eye(3))
    with pytest.raises(NotImplementedError):
        Normal(np.eye(3)) * p1


def test_shorthands():
    model = Graph()
    p = GP(EQ(), graph=model)

    # Construct a normal distribution that serves as in input.
    x = p(1)
    assert isinstance(x, At)
    assert type_parameter(x) is p
    assert x.get() == 1
    assert str(p(x)) == '{}({})'.format(str(p), str(x))
    assert repr(p(x)) == '{}({})'.format(repr(p), repr(x))

    # Construct a normal distribution that does not serve as an input.
    x = Normal(np.ones((1, 1)))
    with pytest.raises(RuntimeError):
        type_parameter(x)
    with pytest.raises(RuntimeError):
        x.get()
    with pytest.raises(RuntimeError):
        p | (x, 1)

    # Test shorthands for stretching and selection.
    p = GP(EQ(), graph=Graph())
    assert str(p > 2) == str(p.stretch(2))
    assert str(p[0]) == str(p.select(0))


def test_properties():
    model = Graph()

    p1 = GP(EQ(), graph=model)
    p2 = GP(EQ().stretch(2), graph=model)
    p3 = GP(EQ().periodic(10), graph=model)

    p = p1 + 2 * p2

    assert p.stationary == True, 'stationary'
    assert p.var == 1 + 2 ** 2, 'var'
    allclose(p.length_scale, (1 + 2 * 2 ** 2) / (1 + 2 ** 2))
    assert p.period == np.inf, 'period'

    assert p3.period == 10, 'period'

    p = p3 + p

    assert p.stationary == True, 'stationary 2'
    assert p.var == 1 + 2 ** 2 + 1, 'var 2'
    assert p.period == np.inf, 'period 2'

    p = p + GP(Linear(), graph=model)

    assert p.stationary == False, 'stationary 3'


def test_marginals():
    model = Graph()
    p = GP(EQ(), TensorProductMean(lambda x: x ** 2), graph=model)
    x = np.linspace(0, 5, 10)

    # Check that `marginals` outputs the right thing.
    mean, lower, upper = p(x).marginals()
    var = B.diag(p.kernel(x))
    allclose(mean, p.mean(x)[:, 0])
    allclose(lower, p.mean(x)[:, 0] - 2 * var ** .5)
    allclose(upper, p.mean(x)[:, 0] + 2 * var ** .5)

    # Test correctness.
    y = p(x).sample()
    mean, lower, upper = (p | (x, y))(x).marginals()
    allclose(mean, y[:, 0])
    assert B.mean(B.abs(upper - lower)) <= 1e-5

    mean, lower, upper = (p | (x, y))(x + 100).marginals()
    allclose(mean, p.mean(x + 100)[:, 0])
    allclose(upper - lower, 4 * np.ones(10))


def test_observations_and_conditioning():
    model = Graph()
    p1 = GP(EQ(), graph=model)
    p2 = GP(EQ(), graph=model)
    p = p1 + p2
    x = np.linspace(0, 5, 10)
    y = p(x).sample()
    y1 = p1(x).sample()

    # Test all ways of conditioning, including shorthands.
    obs1 = Obs(p(x), y)
    obs2 = Obs(x, y, ref=p)
    allclose(obs1.y, obs2.y)
    allclose(obs1.K_x, obs2.K_x)

    obs3 = Obs((p(x), y), (p1(x), y1))
    obs4 = Obs((x, y), (p1(x), y1), ref=p)
    allclose(obs3.y, obs4.y)
    allclose(obs3.K_x, obs4.K_x)

    def assert_equal_mean_var(x, *ys):
        for y in ys:
            allclose(x.mean, y.mean)
            allclose(x.var, y.var)

    assert_equal_mean_var(p.condition(x, y)(x),
                          p.condition(p(x), y)(x),
                          (p | (x, y))(x),
                          (p | (p(x), y))(x),
                          p.condition(obs1)(x),
                          p.condition(obs2)(x),
                          (p | obs1)(x),
                          (p | obs2)(x))

    assert_equal_mean_var(p.condition((x, y), (p1(x), y1))(x),
                          p.condition((p(x), y), (p1(x), y1))(x),
                          (p | [(x, y), (p1(x), y1)])(x),
                          (p | [(p(x), y), (p1(x), y1)])(x),
                          p.condition(obs3)(x),
                          p.condition(obs4)(x),
                          (p | obs3)(x),
                          (p | obs4)(x))

    # Check conditioning multiple processes at once.
    p1_post, p2_post, p_post = (p1, p2, p) | obs1
    p1_post, p2_post, p_post = p1_post(x), p2_post(x), p_post(x)
    p1_post2, p2_post2, p_post2 = (p1 | obs1)(x), (p2 | obs1)(x), (p | obs1)(x)

    allclose(p1_post.mean, p1_post2.mean)
    allclose(p1_post.var, p1_post2.var)
    allclose(p2_post.mean, p2_post2.mean)
    allclose(p2_post.var, p2_post2.var)
    allclose(p_post.mean, p_post2.mean)
    allclose(p_post.var, p_post2.var)

    # Test `At` check.
    with pytest.raises(ValueError):
        Obs(0, 0)
    with pytest.raises(ValueError):
        Obs((0, 0), (0, 0))
    with pytest.raises(ValueError):
        SparseObs(0, p, (0, 0))

    # Test that `Graph.logpdf` takes an `Observations` object.
    obs = Obs(p(x), y)
    assert model.logpdf(obs) == p(x).logpdf(y)


def test_stretching():
    model = Graph()

    # Test construction:
    p = GP(EQ(), TensorProductMean(lambda x: x ** 2), graph=model)
    assert str(p.stretch(1)) == 'GP(EQ() > 1, <lambda> > 1)'

    # Test case:
    p = GP(EQ(), graph=model)
    p2 = p.stretch(5)

    n = 5
    x = np.linspace(0, 10, n)[:, None]
    y = p2(x).sample()

    post = p.condition(p2(x), y)
    allclose(post(x / 5).mean, y)
    assert abs_err(B.diag(post(x / 5).var)) <= 1e-10

    post = p2.condition(p(x), y)
    allclose(post(x * 5).mean, y)
    assert abs_err(B.diag(post(x * 5).var)) <= 1e-10


def test_shifting():
    model = Graph()

    # Test construction:
    p = GP(Linear(), TensorProductMean(lambda x: x ** 2), graph=model)
    assert str(p.shift(1)) == 'GP(Linear() shift 1, <lambda> shift 1)'

    # Test case:
    p = GP(EQ(), graph=model)
    p2 = p.shift(5)

    n = 5
    x = np.linspace(0, 10, n)[:, None]
    y = p2(x).sample()

    post = p.condition(p2(x), y)
    allclose(post(x - 5).mean, y)
    assert abs_err(B.diag(post(x - 5).var)) <= 1e-10

    post = p2.condition(p(x), y)
    allclose(post(x + 5).mean, y)
    assert abs_err(B.diag(post(x + 5).var)) <= 1e-10


def test_input_transform():
    model = Graph()

    # Test construction:
    p = GP(EQ(), TensorProductMean(lambda x: x ** 2), graph=model)
    assert str(p.transform(lambda x: x)) == \
           'GP(EQ() transform <lambda>, <lambda> transform <lambda>)'

    # Test case:
    p = GP(EQ(), graph=model)
    p2 = p.transform(lambda x: x / 5)

    n = 5
    x = np.linspace(0, 10, n)[:, None]
    y = p2(x).sample()

    post = p.condition(p2(x), y)
    allclose(post(x / 5).mean, y)
    assert abs_err(B.diag(post(x / 5).var)) <= 1e-10

    post = p2.condition(p(x), y)
    allclose(post(x * 5).mean, y)
    assert abs_err(B.diag(post(x * 5).var)) <= 1e-10


def test_selection():
    model = Graph()

    # Test construction:
    p = GP(EQ(), TensorProductMean(lambda x: x ** 2), graph=model)
    assert str(p.select(1)) == 'GP(EQ() : [1], <lambda> : [1])'
    assert str(p.select(1, 2)) == 'GP(EQ() : [1, 2], <lambda> : [1, 2])'

    # Test case:
    p = GP(EQ(), graph=model)  # 1D
    p2 = p.select(0)  # 2D

    n = 5
    x = np.linspace(0, 10, n)[:, None]
    x1 = np.concatenate((x, np.random.randn(n, 1)), axis=1)
    x2 = np.concatenate((x, np.random.randn(n, 1)), axis=1)
    y = p2(x).sample()

    post = p.condition(p2(x1), y)
    allclose(post(x).mean, y)
    assert abs_err(B.diag(post(x).var)) <= 1e-10

    post = p.condition(p2(x2), y)
    allclose(post(x).mean, y)
    assert abs_err(B.diag(post(x).var)) <= 1e-10

    post = p2.condition(p(x), y)
    allclose(post(x1).mean, y)
    allclose(post(x2).mean, y)
    assert abs_err(B.diag(post(x1).var)) <= 1e-10
    assert abs_err(B.diag(post(x2).var)) <= 1e-10


def test_derivative():
    # Test construction:
    p = GP(EQ(), TensorProductMean(lambda x: x ** 2), graph=Graph())
    assert str(p.diff(1)) == 'GP(d(1) EQ(), d(1) <lambda>)'

    # Test case:
    model = Graph()
    x = B.linspace(tf.float64, 0, 1, 100)[:, None]
    y = 2 * x

    p = GP(EQ(), graph=model)
    dp = p.diff()

    # Test conditioning on function.
    assert abs_err(dp.condition(p(x), y)(x).mean, 2) <= 1e-3

    # Test conditioning on derivative.
    post = p.condition((B.cast(tf.float64, 0),
                        B.cast(tf.float64, 0)), (dp(x), y))
    assert abs_err(post(x).mean, x ** 2) <= 1e-3


def test_multi_sample():
    model = Graph()
    p1 = GP(0, 1, graph=model)
    p2 = GP(0, 2, graph=model)
    p3 = GP(0, 3, graph=model)

    x1 = np.linspace(0, 1, 10)
    x2 = np.linspace(0, 1, 20)
    x3 = np.linspace(0, 1, 30)

    s1, s2, s3 = model.sample(p1(x1), p2(x2), p3(x3))

    assert s1.shape == (10, 1)
    assert s2.shape == (20, 1)
    assert s3.shape == (30, 1)
    assert np.shape(model.sample(p1(x1))) == (10, 1)

    assert abs_err(s1 - 1) <= 1e-4
    assert abs_err(s2 - 2) <= 1e-4
    assert abs_err(s3 - 3) <= 1e-4


def test_multi_conditioning():
    model = Graph()

    p1 = GP(EQ(), graph=model)
    p2 = GP(2 * Exp().stretch(2), graph=model)
    p3 = GP(.5 * RQ(1e-1).stretch(.5), graph=model)

    p = p1 + p2 + p3

    x1 = np.linspace(0, 2, 10)
    x2 = np.linspace(1, 3, 10)
    x3 = np.linspace(0, 3, 10)

    s1, s2 = model.sample(p1(x1), p2(x2))

    post1 = ((p | (p1(x1), s1)) | ((p2 | (p1(x1), s1))(x2), s2))(x3)
    post2 = (p | ((p1(x1), s1), (p2(x2), s2)))(x3)
    post3 = (p | ((p2(x2), s2), (p1(x1), s1)))(x3)

    allclose(post1.mean, post2.mean, desc='means 1', atol=1e-6, rtol=1e-6)
    allclose(post1.mean, post3.mean, desc='means 2', atol=1e-6, rtol=1e-6)
    allclose(post1.var, post2.var)
    allclose(post1.var, post3.var)


def test_multi_logpdf():
    model = Graph()

    p1 = GP(EQ(), graph=model)
    p2 = GP(2 * Exp(), graph=model)
    p3 = p1 + p2

    x1 = np.linspace(0, 2, 5)
    x2 = np.linspace(1, 3, 6)
    x3 = np.linspace(2, 4, 7)

    y1, y2, y3 = model.sample(p1(x1), p2(x2), p3(x3))

    # Test case that only one process is fed.
    allclose(p1(x1).logpdf(y1), model.logpdf(p1(x1), y1))
    allclose(p1(x1).logpdf(y1), model.logpdf((p1(x1), y1)))

    # Test that all inputs must be specified.
    with pytest.raises(ValueError):
        model.logpdf((x1, y1), (p2(x2), y2))
    with pytest.raises(ValueError):
        model.logpdf((p1(x1), y1), (x2, y2))

    # Compute the logpdf with the product rule.
    logpdf1 = p1(x1).logpdf(y1) + p2(x2).logpdf(y2) + \
              (p3 | ((p1(x1), y1), (p2(x2), y2)))(x3).logpdf(y3)
    logpdf2 = model.logpdf((p1(x1), y1), (p2(x2), y2), (p3(x3), y3))
    allclose(logpdf1, logpdf2)


def test_approximate_multiplication():
    model = Graph()

    # Construct model.
    p1 = GP(EQ(), 20, graph=model)
    p2 = GP(EQ(), 20, graph=model)
    p_prod = p1 * p2
    x = np.linspace(0, 10, 50)

    # Sample functions.
    s1, s2 = model.sample(p1(x), p2(x))

    # Infer product.
    post = p_prod.condition((p1(x), s1), (p2(x), s2))
    assert rel_err(post(x).mean, s1 * s2) <= 1e-2

    # Perform division.
    cur_epsilon = B.epsilon
    B.epsilon = 1e-8
    post = p2.condition((p1(x), s1), (p_prod(x), s1 * s2))
    assert rel_err(post(x).mean, s2) <= 1e-2
    B.epsilon = cur_epsilon

    # Check graph check.
    model2 = Graph()
    p3 = GP(EQ(), graph=model2)
    with pytest.raises(RuntimeError):
        p3 * p1


def test_naming():
    model = Graph()

    p1 = GP(EQ(), 1, graph=model)
    p2 = GP(EQ(), 2, graph=model)

    # Test setting and getting names.
    p1.name = 'name'

    assert model['name'] is p1
    assert p1.name == 'name'
    assert model[p1] == 'name'
    with pytest.raises(KeyError):
        model['other_name']
    with pytest.raises(KeyError):
        model[p2]

    # Check that names can not be doubly assigned.
    def doubly_assign():
        p2.name = 'name'

    with pytest.raises(RuntimeError):
        doubly_assign()

    # Move name to other GP.
    p1.name = 'other_name'
    p2.name = 'name'

    # Check that everything has been properly assigned.
    assert model['name'] is p2
    assert p2.name == 'name'
    assert model[p2] == 'name'
    assert model['other_name'] is p1
    assert p1.name == 'other_name'
    assert model[p1] == 'other_name'

    # Test giving a name to the constructor.
    p3 = GP(EQ(), name='yet_another_name', graph=model)
    assert model['yet_another_name'] is p3
    assert p3.name == 'yet_another_name'
    assert model[p3] == 'yet_another_name'


def test_formatting():
    p = 2 * GP(EQ(), 1, graph=Graph())
    assert str(p.display(lambda x: x ** 2)) == 'GP(16 * EQ(), 4 * 1)'


def test_sparse_conditioning():
    model = Graph()
    f = GP(EQ().stretch(3), graph=model)
    e = GP(1e-2 * Delta(), graph=model)
    x = np.linspace(0, 5, 10)
    x_new = np.linspace(6, 10, 10)

    y = f(x).sample()

    # Test that noise matrix must indeed be diagonal.
    with pytest.raises(RuntimeError):
        SparseObs(f(x), f, f(x), y).elbo

    # Test posterior.
    post_sparse = (f | SparseObs(f(x), e, f(x), y))(x_new)
    post_ref = (f | ((f + e)(x), y))(x_new)
    allclose(post_sparse.mean, post_ref.mean, desc='means 1', atol=1e-6,
             rtol=1e-6)
    allclose(post_sparse.var, post_ref.var)

    post_sparse = (f | SparseObs(f(x), e, (2 * f + 2)(x), 2 * y + 2))(x_new)
    post_ref = (f | ((2 * f + 2 + e)(x), 2 * y + 2))(x_new)
    allclose(post_sparse.mean, post_ref.mean, desc='means 2', atol=1e-6,
             rtol=1e-6)
    allclose(post_sparse.var, post_ref.var)

    post_sparse = (f | SparseObs((2 * f + 2)(x), e, f(x), y))(x_new)
    post_ref = (f | ((f + e)(x), y))(x_new)
    allclose(post_sparse.mean, post_ref.mean, desc='means 3', atol=1e-6,
             rtol=1e-6)
    allclose(post_sparse.var, post_ref.var)

    # Test ELBO.
    e = GP(1e-2 * Delta(), graph=model)
    allclose(SparseObs(f(x), e, f(x), y).elbo, (f + e)(x).logpdf(y))
    allclose(SparseObs(f(x), e, (2 * f + 2)(x), 2 * y + 2).elbo,
             (2 * f + 2 + e)(x).logpdf(2 * y + 2))
    allclose(SparseObs((2 * f + 2)(x), e, f(x), y).elbo, (f + e)(x).logpdf(y))

    # Test multiple observations.
    x1 = np.linspace(0, 5, 10)
    x2 = np.linspace(10, 15, 10)
    x_new = np.linspace(6, 9, 10)
    x_ind = np.concatenate((x1, x2, x_new), axis=0)
    y1, y2 = model.sample((f + e)(x1), (f + e)(x2))

    post_sparse = (f | SparseObs(f(x_ind),
                                 (e, f(Unique(x1)), y1),
                                 (e, f(Unique(x2)), y2)))(x_new)
    post_ref = (f | Obs(((f + e)(x1), y1), ((f + e)(x2), y2)))(x_new)
    allclose(post_sparse.mean, post_ref.mean)
    allclose(post_sparse.var, post_ref.var)

    # Test multiple observations and multiple inducing points.
    post_sparse = (f | SparseObs((f(x1), f(x2), f(x_new)),
                                 (e, f(Unique(x1)), y1),
                                 (e, f(Unique(x2)), y2)))(x_new)
    allclose(post_sparse.mean, post_ref.mean, desc='means 4', atol=1e-6,
             rtol=1e-6)
    allclose(post_sparse.var, post_ref.var)

    # Test multiple inducing points.
    x = np.linspace(0, 5, 10)
    x_new = np.linspace(6, 10, 10)
    x_ind1 = x[:5]
    x_ind2 = x[5:]
    y = model.sample((f + e)(x))

    post_sparse = (f | SparseObs((f(x_ind1), f(x_ind2)), e, f(x), y))(x_new)
    post_ref = (f | ((f + e)(x), y))(x_new)
    allclose(post_sparse.mean, post_ref.mean, desc='means 5', atol=1e-4,
             rtol=1e-4)
    allclose(post_sparse.var, post_ref.var)

    # Test caching of mean.
    obs = SparseObs(f(x), e, f(x), y)
    mu = obs.mu
    allclose(mu, obs.mu)

    # Test caching of corrective kernel parameter.
    obs = SparseObs(f(x), e, f(x), y)
    A = obs.A
    allclose(A, obs.A)

    # Test caching of elbo.
    obs = SparseObs(f(x), e, f(x), y)
    elbo = obs.elbo
    allclose(elbo, obs.elbo)

    # Test that `Graph.logpdf` takes an `SparseObservations` object.
    obs = SparseObs(f(x), e, f(x), y)
    allclose(model.logpdf(obs), (f + e)(x).logpdf(y))


def test_case_summation_with_itself():
    # Test summing the same GP with itself.
    model = Graph()
    p1 = GP(EQ(), graph=model)
    p2 = p1 + p1 + p1 + p1 + p1
    x = np.linspace(0, 10, 5)[:, None]

    allclose(p2(x).var, 25 * p1(x).var)
    allclose(p2(x).mean, np.zeros((5, 1)))

    y = np.random.randn(5, 1)
    allclose(p2.condition(p1(x), y)(x).mean, 5 * y)


def test_case_additive_model():
    # Test some additive model.

    model = Graph()
    p1 = GP(EQ(), graph=model)
    p2 = GP(EQ(), graph=model)
    p3 = p1 + p2

    n = 5
    x = np.linspace(0, 10, n)[:, None]
    y1 = p1(x).sample()
    y2 = p2(x).sample()

    # First, test independence:
    allclose(model.kernels[p2, p1](x), np.zeros((n, n)))
    allclose(model.kernels[p1, p2](x), np.zeros((n, n)))

    # Now run through some test cases:

    obs = Obs(p1(x), y1)
    post = (p3 | obs) | ((p2 | obs)(x), y2)
    allclose(post(x).mean, y1 + y2)

    obs = Obs(p2(x), y2)
    post = (p3 | obs) | ((p1 | obs)(x), y1)
    allclose(post(x).mean, y1 + y2)

    obs = Obs(p1(x), y1)
    post = (p2 | obs) | ((p3 | obs)(x), y1 + y2)
    allclose(post(x).mean, y2)

    obs = Obs(p3(x), y1 + y2)
    post = (p2 | obs) | ((p1 | obs)(x), y1)
    allclose(post(x).mean, y2)

    allclose(p3.condition(x, y1 + y2)(x).mean, y1 + y2)


def test_case_fd_derivative():
    x = np.linspace(0, 10, 50)[:, None]
    y = np.sin(x)

    model = Graph()
    p = GP(.7 * EQ().stretch(1.), graph=model)
    dp = (p.shift(-1e-3) - p.shift(1e-3)) / 2e-3

    assert abs_err(np.cos(x) - dp.condition(p(x), y)(x).mean) <= 1e-4


def test_case_reflection():
    model = Graph()
    p = GP(EQ(), graph=model)
    p2 = 5 - p

    x = np.linspace(0, 1, 10)[:, None]
    y = p(x).sample()

    assert abs_err(p2.condition(p(x), y)(x).mean - (5 - y)) <= 1e-5
    assert abs_err(p.condition(p2(x), 5 - y)(x).mean - y) <= 1e-5

    model = Graph()
    p = GP(EQ(), graph=model)
    p2 = -p

    x = np.linspace(0, 1, 10)[:, None]
    y = p(x).sample()

    assert abs_err(p2.condition(p(x), y)(x).mean + y) <= 1e-5
    assert abs_err(p.condition(p2(x), -y)(x).mean - y) <= 1e-5


def test_case_approximate_derivative():
    model = Graph()
    x = np.linspace(0, 1, 100)[:, None]
    y = 2 * x

    p = GP(EQ().stretch(1.), graph=model)
    dp = p.diff_approx()

    # Test conditioning on function.
    assert abs_err(dp.condition(p(x), y)(x).mean - 2) <= 1e-3

    # Add some regularisation for this test case.
    orig_epsilon = B.epsilon
    B.epsilon = 1e-10

    # Test conditioning on derivative.
    post = p.condition((0, 0), (dp(x), y))
    assert abs_err(post(x).mean - x ** 2) <= 1e-3

    # Set regularisation back.
    B.epsilon = orig_epsilon


def test_case_blr():
    model = Graph()
    x = np.linspace(0, 10, 100)

    slope = GP(1, graph=model)
    intercept = GP(1, graph=model)
    f = slope * (lambda x: x) + intercept
    y = f + 1e-2 * GP(Delta(), graph=model)

    # Sample observations, true slope, and intercept.
    y_obs, true_slope, true_intercept = \
        model.sample(y(x), slope(0), intercept(0))

    # Predict.
    post_slope, post_intercept = (slope, intercept) | Obs(y(x), y_obs)
    mean_slope, mean_intercept = post_slope(0).mean, post_intercept(0).mean

    assert np.abs(true_slope[0, 0] - mean_slope[0, 0]) <= 5e-2
    assert np.abs(true_intercept[0, 0] - mean_intercept[0, 0]) <= 5e-2



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

import logging

import tensorflow as tf
from lab import B
from plum import Dispatcher, Self, Referentiable

from .field import get_field
from .function_field import (
    StretchedFunction,
    ShiftedFunction,
    SelectedFunction,
    InputTransformedFunction,
    DerivativeFunction,
    TensorProductFunction,
    ZeroFunction,
    OneFunction,
    ScaledFunction,
    ProductFunction,
    SumFunction,
    Function
)
from .input import Input
from .matrix import dense
from .util import uprank

__all__ = ['TensorProductMean', 'DerivativeMean']

log = logging.getLogger(__name__)


class Mean(Function, Referentiable):
    """Mean function.

    Means can be added and multiplied.
    """

    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(object)
    def __call__(self, x):
        """Construct the mean for a design matrix.

        Args:
            x (input): Points to construct the mean at.

        Returns:
            tensor: Mean vector as a rank 2 column vector.
        """
        raise NotImplementedError()

    @_dispatch(Input)
    def __call__(self, x):
        # This should not have been reached. Attempt to unwrap.
        return self(x.get())


# Register the field.
@get_field.extend(Mean)
def get_field(a): return Mean


class SumMean(Mean, SumFunction, Referentiable):
    """Sum of two means."""

    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(object)
    def __call__(self, x):
        return B.add(self[0](x), self[1](x))


class ProductMean(Mean, ProductFunction, Referentiable):
    """Product of two means."""

    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(object)
    def __call__(self, x):
        return B.multiply(self[0](x), self[1](x))


class ScaledMean(Mean, ScaledFunction, Referentiable):
    """Scaled mean."""

    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(object)
    def __call__(self, x):
        return B.multiply(self.scale, self[0](x))


class StretchedMean(Mean, StretchedFunction, Referentiable):
    """Stretched mean."""

    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(object)
    @uprank
    def __call__(self, x):
        return self[0](B.divide(x, self.stretches[0]))


class ShiftedMean(Mean, ShiftedFunction, Referentiable):
    """Shifted mean."""

    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(object)
    @uprank
    def __call__(self, x):
        return self[0](B.subtract(x, self.shifts[0]))


class SelectedMean(Mean, SelectedFunction, Referentiable):
    """Mean with particular input dimensions selected."""

    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(object)
    @uprank
    def __call__(self, x):
        return self[0](B.take(x, self.dims[0], axis=1))


class InputTransformedMean(Mean, InputTransformedFunction, Referentiable):
    """Input-transformed mean."""

    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(object)
    def __call__(self, x):
        return self[0](uprank(self.fs[0](x)))


class OneMean(Mean, OneFunction, Referentiable):
    """Constant mean of `1`."""

    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(B.Numeric)
    @uprank
    def __call__(self, x):
        return B.ones(B.dtype(x), B.shape(x)[0], 1)


class ZeroMean(Mean, ZeroFunction, Referentiable):
    """Constant mean of `0`."""

    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(B.Numeric)
    @uprank
    def __call__(self, x):
        return B.zeros(B.dtype(x), B.shape(x)[0], 1)


class TensorProductMean(Mean, TensorProductFunction, Referentiable):
    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(B.Numeric)
    @uprank
    def __call__(self, x):
        return uprank(self.fs[0](x))


class DerivativeMean(Mean, DerivativeFunction, Referentiable):
    """Derivative of mean."""
    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(B.Numeric)
    @uprank
    def __call__(self, x):
        i = self.derivs[0]
        with tf.GradientTape() as t:
            xi = x[:, i:i + 1]
            t.watch(xi)
            x = B.concat(x[:, :i], xi, x[:, i + 1:], axis=1)
            out = dense(self[0](x))
            return t.gradient(out, xi, unconnected_gradients='zero')


class PosteriorMean(Mean, Referentiable):
    """Posterior mean.

    Args:
        m_i (:class:`.mean.Mean`): Mean of process corresponding to
            the input.
        m_z (:class:`.mean.Mean`): Mean of process corresponding to
            the data.
        k_zi (:class:`.kernel.Kernel`): Kernel between processes
            corresponding to the data and the input respectively.
        z (input): Locations of data.
        K_z (:class:`.matrix.Dense`): Kernel matrix of data.
        y (tensor): Observations to condition on.
    """

    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, m_i, m_z, k_zi, z, K_z, y):
        self.m_i = m_i
        self.m_z = m_z
        self.k_zi = k_zi
        self.z = z
        self.K_z = K_z
        self.y = uprank(y)

    @_dispatch(object)
    def __call__(self, x):
        diff = B.subtract(self.y, self.m_z(self.z))
        return B.add(self.m_i(x),
                     B.qf(self.K_z, self.k_zi(self.z, x), diff))



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

import operator
from types import FunctionType as PythonFunction
import numpy as np

from lab import B
from plum import Dispatcher, Referentiable, Self
from stheno.field import squeeze, mul, add, SumElement, ProductElement, \
    ScaledElement, OneElement, ZeroElement, WrappedElement, JoinElement, \
    Formatter, priority

from .field import Element, new, get_field, broadcast

__all__ = []

_dispatch = Dispatcher()


@_dispatch(B.Numeric)
def _shape(x):
    return B.shape(x)


@_dispatch(object)
def _shape(x):
    return np.shape(x)


@_dispatch(tuple, tuple)
def tuple_equal(x, y):
    """Check tuples for equality.

    Args:
        x (tuple): First tuple.
        y (tuple): Second tuple.

    Returns:
        bool: `x` and `y` are equal.
    """
    return len(x) == len(y) and \
           all([_shape(xi) == _shape(yi) and B.all(xi == yi)
                for xi, yi in zip(x, y)])


@_dispatch(B.Numeric)
def to_tensor(x):
    """Convert object to tensor.

    Args:
        x (object): Object to convert to tensor.

    Returns:
        tensor: `x` as a tensor.
    """
    return x


@_dispatch({tuple, list})
def to_tensor(x):
    return B.stack(*x, axis=0)


class Function(Element, Referentiable):
    """A function.

    Crucially, this is not a field, so that it can be inherited.
    """
    _dispatch = Dispatcher(in_class=Self)

    def stretch(self, *stretches):
        """Stretch the function.

        Args:
            *stretches (tensor): Per input, extent to stretch by.

        Returns:
            :class:`.function_field.Function`: Stretched function.
        """
        return stretch(self, *stretches)

    def __gt__(self, stretch):
        """Shorthand for :meth:`.function_field.Function.stretch`."""
        return self.stretch(stretch)

    def shift(self, *amounts):
        """Shift the inputs of an function by a certain amount.

        Args:
            *amounts (tensor): Per input, amount to shift by.

        Returns:
            :class:`.function_field.Function`: Shifted function.
        """
        return shift(self, *amounts)

    def select(self, *dims):
        """Select particular dimensions of the input features.

        Args:
            *dims (int, sequence, or None): Per input, dimensions to select.
                Set to `None` to select all.

        Returns:
            :class:`.function_field.Function`: Function with dimensions of the
                input features selected.
        """
        return select(self, *dims)

    def transform(self, *fs):
        """Transform the inputs of a function.

        Args:
            *fs (int or tuple): Per input, transformation. Set to `None` to
                not perform a transformation.

        Returns:
            :class:`.function_field.Function`: Function with its inputs
                transformed.
        """
        return transform(self, *fs)

    def diff(self, *derivs):
        """Differentiate a function.

        Args:
            *derivs (int): Per input, dimension of the feature which to take
                the derivatives with respect to. Set to `None` to not take a
                derivative.

        Returns:
            :class:`.function_field.Function`: Derivative of the Function.
        """
        return differentiate(self, *derivs)


# Register the field.
@get_field.extend(Function)
def get_field(a): return Function


class OneFunction(Function, OneElement):
    """The constant function `1`."""


class ZeroFunction(Function, ZeroElement):
    """The constant function `0`."""


class WrappedFunction(Function, WrappedElement):
    """A wrapped function."""


class JoinFunction(Function, JoinElement):
    """Two wrapped functions."""


class SumFunction(Function, SumElement):
    """A sum of two functions."""


class ScaledFunction(Function, ScaledElement):
    """A scaled function."""


class ProductFunction(Function, ProductElement):
    """A product of two functions."""


class StretchedFunction(WrappedFunction, Referentiable):
    """Stretched function.

    Args:
        e (:class:`.function_field.Function`): Function to stretch.
        *stretches (tensor): Extent of stretches.
    """
    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, e, *stretches):
        WrappedFunction.__init__(self, e)
        self.stretches = tuple(to_tensor(x) for x in stretches)

    @_dispatch(object, Formatter)
    def display(self, e, formatter):
        stretches = tuple(formatter(s) for s in self.stretches)
        return '{} > {}'.format(e, squeeze(stretches))

    @_dispatch(Self)
    def __eq__(self, other):
        return self[0] == other[0] and \
               tuple_equal(self.stretches, other.stretches)


class ShiftedFunction(WrappedFunction, Referentiable):
    """Shifted function.

    Args:
        e (:class:`.function_field.Function`): Function to shift.
        *shifts (tensor): Shift amounts.
    """
    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, e, *shifts):
        WrappedFunction.__init__(self, e)
        self.shifts = tuple(to_tensor(x) for x in shifts)

    @_dispatch(object, Formatter)
    def display(self, e, formatter):
        shifts = tuple(formatter(s) for s in self.shifts)
        return '{} shift {}'.format(e, squeeze(shifts))

    @_dispatch(Self)
    def __eq__(self, other):
        return self[0] == other[0] and \
               tuple_equal(self.shifts, other.shifts)


@_dispatch({tuple, list})
def _to_list(x):
    return list(x)


@_dispatch(object)
def _to_list(x):
    if B.rank(x) == 0:
        return [x]
    elif B.rank(x) == 1:
        return x
    else:
        raise ValueError('Could not convert "{}" to a list.')


class SelectedFunction(WrappedFunction, Referentiable):
    """Select particular dimensions of the input features.

    Args:
        e (:class:`.function_field.Function`): Function to wrap.
        *dims (tensor): Dimensions to select.
    """
    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, e, *dims):
        WrappedFunction.__init__(self, e)
        self.dims = tuple(None if x is None else _to_list(x) for x in dims)

    @_dispatch(object, Formatter)
    def display(self, e, formatter):
        return '{} : {}'.format(e, squeeze(tuple(self.dims)))

    @_dispatch(Self)
    def __eq__(self, other):
        return self[0] == other[0] and \
               tuple_equal(self.dims, other.dims)


class InputTransformedFunction(WrappedFunction, Referentiable):
    """Transform inputs of a function.

    Args:
        e (:class:`.function_field.Function`): Function to wrap.
        *fs (tensor): Per input, the transformation. Set to `None` to not
            do a transformation.
    """
    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, e, *fs):
        WrappedFunction.__init__(self, e)
        self.fs = fs

    @_dispatch(object, Formatter)
    def display(self, e, formatter):
        # Safely get a function's name.
        def name(f):
            return 'None' if f is None else f.__name__

        if len(self.fs) == 1:
            fs = name(self.fs[0])
        else:
            fs = '({})'.format(', '.join(name(f) for f in self.fs))
        return '{} transform {}'.format(e, fs)

    @_dispatch(Self)
    def __eq__(self, other):
        return self[0] == other[0] and \
               tuple_equal(self.fs, other.fs)


class DerivativeFunction(WrappedFunction, Referentiable):
    """Compute the derivative of a function.

    Args:
        e (:class:`.field_function.Element`): Function to compute the
            derivative of.
        *derivs (tensor): Per input, the index of the dimension which to
            take the derivative of. Set to `None` to not take a derivative.
    """
    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, e, *derivs):
        WrappedFunction.__init__(self, e)
        self.derivs = derivs

    @_dispatch(object, Formatter)
    def display(self, e, formatter):
        if len(self.derivs) == 1:
            derivs = '({})'.format(self.derivs[0])
        else:
            derivs = self.derivs
        return 'd{} {}'.format(derivs, e)

    @_dispatch(Self)
    def __eq__(self, other):
        return self[0] == other[0] and \
               tuple_equal(self.derivs, other.derivs)


class TensorProductFunction(Function, Referentiable):
    """An element built from a product of functions for each input.

    Args:
        *fs (function): Per input, a function.
    """
    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, *fs):
        self.fs = fs

    @_dispatch(Formatter)
    def display(self, formatter):
        if len(self.fs) == 1:
            return self.fs[0].__name__
        else:
            return '({})'.format(' x '.join(f.__name__ for f in self.fs))

    @_dispatch(Self)
    def __eq__(self, other):
        return tuple_equal(self.fs, other.fs)


@_dispatch(object, [object])
def stretch(a, *stretches):
    """Stretch a function.

    Args:
        a (:class:`.function_field.Function`): Function to stretch.
        *stretches (tensor): Per input, extent of stretches.

    Returns:
        :class:`.function_field.Function`: Stretched function.
    """
    raise NotImplementedError('Stretching not implemented for "{}".'
                              ''.format(type(a).__name__))


@_dispatch(object, [object])
def shift(a, *shifts):
    """Shift a function.

    Args:
        a (:class:`.function_field.Function`): Function to shift.
        *shifts (tensor): Per input, amount of shift.

    Returns:
        :class:`.function_field.Function`: Shifted element.
    """
    raise NotImplementedError('Shifting not implemented for "{}".'
                              ''.format(type(a).__name__))


@_dispatch(object, [object])
def select(a, *dims):
    """Select dimensions from the inputs.

    Args:
        a (:class:`.function_field.Function`): Function  to wrap.
        *dims (int): Per input, dimensions to select. Set to `None` to select
            all.

    Returns:
        :class:`.function_field.Function`: Function with particular dimensions
            from the inputs selected.
    """
    raise NotImplementedError('Selection not implemented for "{}".'
                              ''.format(type(a).__name__))


@_dispatch(object, [object])
def transform(a, *fs):
    """Transform the inputs of a function.

    Args:
        a (:class:`.function_field.Function`): Function to wrap.
        *fs (int): Per input, the transform. Set to `None` to not perform a
            transform.

    Returns:
        :class:`.function_field.Function`: Function with its inputs
            transformed.
    """
    raise NotImplementedError('Input transforms not implemented for "{}".'
                              ''.format(type(a).__name__))


@_dispatch(object, [object])
def differentiate(a, *derivs):
    """Differentiate a function.

    Args:
        a (:class:`.function_field.Function`): Function to differentiate.
        *derivs (int): Per input, dimension of the feature which to take
            the derivatives with respect to. Set to `None` to not take a
            derivative.

    Returns:
        :class:`.function_field.Function`: Derivative of the function.
    """
    raise NotImplementedError('Differentiation not implemented for "{}".'
                              ''.format(type(a).__name__))


# Handle conversion of Python functions.

@mul.extend(Element, PythonFunction, precedence=priority)
def mul(a, b): return mul(a, new(a, TensorProductFunction)(b))


@mul.extend(PythonFunction, Element, precedence=priority)
def mul(a, b): return mul(new(b, TensorProductFunction)(a), b)


@add.extend(Element, PythonFunction, precedence=priority)
def add(a, b): return add(a, new(a, TensorProductFunction)(b))


@add.extend(PythonFunction, Element, precedence=priority)
def add(a, b): return add(new(b, TensorProductFunction)(a), b)


# Stretch:

@_dispatch(Element, [object])
def stretch(a, *stretches): return new(a, StretchedFunction)(a, *stretches)


@_dispatch({ZeroElement, OneElement}, [object])
def stretch(a, *stretches): return a


@_dispatch(StretchedFunction, [object])
def stretch(a, *stretches):
    return stretch(a[0], *broadcast(operator.mul, a.stretches, stretches))


# Shifting:

@_dispatch(Element, [object])
def shift(a, *shifts): return new(a, ShiftedFunction)(a, *shifts)


@_dispatch({ZeroElement, OneElement}, [object])
def shift(a, *shifts): return a


@_dispatch(ShiftedFunction, [object])
def shift(a, *shifts):
    return shift(a[0], *broadcast(operator.add, a.shifts, shifts))


# Selection:

@_dispatch(Element, [object])
def select(a, *dims): return new(a, SelectedFunction)(a, *dims)


@_dispatch({ZeroElement, OneElement}, [object])
def select(a, *dims): return a


# Input transforms:

@_dispatch(Element, [object])
def transform(a, *fs): return new(a, InputTransformedFunction)(a, *fs)


@_dispatch({ZeroElement, OneElement}, [object])
def transform(a, *fs): return a


# Differentiation:

@_dispatch(Element, [object])
def differentiate(a, *derivs): return new(a, DerivativeFunction)(a, *derivs)


@_dispatch(ZeroElement, [object])
def differentiate(a, *derivs): return a


@_dispatch(OneElement, [object])
def differentiate(a, *derivs): return new(a, ZeroElement)()



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

from functools import wraps
import logging
from numbers import Number
from time import time
from types import FunctionType

from lab import B
from plum import Dispatcher, Self, Referentiable

__all__ = []

_dispatch = Dispatcher()


@_dispatch(object)
def uprank(x):
    """Ensure that the rank of `x` is 2.

    Args:
        x (tensor): Tensor to uprank.

    Returns:
        tensor: `x` with rank at least 2.
    """
    # Simply return non-numerical inputs.
    if not isinstance(x, B.Numeric):
        return x

    # Now check the rank of `x` and act accordingly.
    rank = B.rank(x)
    if rank > 2:
        raise ValueError('Input must be at most rank 2.')
    elif rank == 2:
        return x
    elif rank == 1:
        return B.expand_dims(x, axis=1)
    else:
        # Rank must be 0.
        return B.expand_dims(B.expand_dims(x, axis=0), axis=1)


@_dispatch(FunctionType)
def uprank(f):
    """A decorator to ensure that the rank of the arguments is two."""

    @wraps(f)
    def wrapped_f(*args):
        return f(*[uprank(x) for x in args])

    return wrapped_f



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

import logging
import operator

import numpy as np
import tensorflow as tf
from lab import B
from plum import Dispatcher, Self, Referentiable, convert

from stheno.function_field import (
    StretchedFunction,
    ShiftedFunction,
    SelectedFunction,
    InputTransformedFunction,
    DerivativeFunction,
    TensorProductFunction,
    stretch,
    transform,
    Function,
    ZeroFunction,
    OneFunction,
    ScaledFunction,
    ProductFunction,
    SumFunction,
    WrappedFunction,
    JoinFunction,
    shift,
    select,
    to_tensor,
    tuple_equal
)
from .field import add, mul, broadcast, get_field, Formatter, need_parens
from .input import Input, Unique
from .matrix import (
    Dense,
    LowRank,
    UniformlyDiagonal,
    Diagonal,
    One,
    Zero,
    dense,
    matrix
)
from .util import uprank

__all__ = ['Kernel',
           'OneKernel',
           'ZeroKernel',
           'ScaledKernel',
           'EQ',
           'RQ',
           'Matern12',
           'Exp',
           'Matern32',
           'Matern52',
           'Delta',
           'FixedDelta',
           'Linear',
           'DerivativeKernel',
           'DecayingKernel',
           'LogKernel']

log = logging.getLogger(__name__)

_dispatch = Dispatcher()


def expand(xs):
    """Expand a sequence to the same element repeated twice if there is only
    one element.

    Args:
        xs (sequence): Sequence to expand.

    Returns:
        object: `xs * 2` or `xs`.
    """
    return xs * 2 if len(xs) == 1 else xs


class Kernel(Function, Referentiable):
    """Kernel function.

    Kernels can be added and multiplied.
    """
    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(object, object)
    def __call__(self, x, y):
        """Construct the kernel matrix between all `x` and `y`.

        Args:
            x (input): First argument.
            y (input, optional): Second argument. Defaults to first
                argument.

        Returns:
            :class:`.matrix.Dense:: Kernel matrix.
        """
        raise RuntimeError('For kernel "{}", could not resolve '
                           'arguments "{}" and "{}".'.format(self, x, y))

    @_dispatch(object)
    def __call__(self, x):
        return self(x, x)

    @_dispatch(Input, Input)
    def __call__(self, x, y):
        # Both input types were not used. Unwrap.
        return self(x.get(), y.get())

    @_dispatch(Input, object)
    def __call__(self, x, y):
        # Left input type was not used. Unwrap.
        return self(x.get(), y)

    @_dispatch(object, Input)
    def __call__(self, x, y):
        # Right input type was not used. Unwrap.
        return self(x, y.get())

    @_dispatch(object, object)
    def elwise(self, x, y):
        """Construct the kernel vector `x` and `y` element-wise.

        Args:
            x (input): First argument.
            y (input, optional): Second argument. Defaults to first
                argument.

        Returns:
            tensor: Kernel vector as a rank 2 column vector.
        """
        # TODO: throw warning
        return B.expand_dims(B.diag(self(x, y)), axis=1)

    @_dispatch(object)
    def elwise(self, x):
        return self.elwise(x, x)

    @_dispatch(Input, Input)
    def elwise(self, x, y):
        # Both input types were not used. Unwrap.
        return self.elwise(x.get(), y.get())

    @_dispatch(Input, object)
    def elwise(self, x, y):
        # Left input type as not used. Unwrap.
        return self.elwise(x.get(), y)

    @_dispatch(object, Input)
    def elwise(self, x, y):
        # Right input type was not used. Unwrap.
        return self.elwise(x, y.get())

    def periodic(self, period=1):
        """Map to a periodic space.

        Args:
            period (tensor, optional): Period. Defaults to `1`.

        Returns:
            :class:`.kernel.Kernel`: Periodic version of the kernel.
        """
        return periodicise(self, period)

    def __reversed__(self):
        """Reverse the arguments of the kernel."""
        return reverse(self)

    @_dispatch(int)
    def __pow__(self, power, modulo=None):
        if power < 0:
            raise ValueError('Cannot raise to a negative power.')
        elif power == 0:
            return 1
        else:
            k = self
            for _ in range(power - 1):
                k *= self
        return k

    @property
    def stationary(self):
        """Stationarity of the kernel."""
        try:
            return self._stationary_cache
        except AttributeError:
            self._stationary_cache = self._stationary
            return self._stationary_cache

    @property
    def _stationary(self):
        return False

    @property
    def var(self):
        """Variance of the kernel."""
        raise RuntimeError('The variance of "{}" could not be determined.'
                           ''.format(self.__class__.__name__))

    @property
    def length_scale(self):
        """Approximation of the length scale of the kernel."""
        raise RuntimeError('The length scale of "{}" could not be determined.'
                           ''.format(self.__class__.__name__))

    @property
    def period(self):
        """Period of the kernel."""
        raise RuntimeError('The period of "{}" could not be determined.'
                           ''.format(self.__class__.__name__))


# Register the field.
@get_field.extend(Kernel)
def get_field(a): return Kernel


class OneKernel(Kernel, OneFunction, Referentiable):
    """Constant kernel of `1`."""

    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(B.Numeric, B.Numeric)
    def __call__(self, x, y):
        if x is y:
            return One(B.dtype(x), B.shape(uprank(x))[0])
        else:
            return One(B.dtype(x),
                       B.shape(uprank(x))[0],
                       B.shape(uprank(y))[0])

    @_dispatch(B.Numeric, B.Numeric)
    @uprank
    def elwise(self, x, y):
        return B.ones(B.dtype(x), B.shape(x)[0], 1)

    @property
    def _stationary(self):
        return True

    @property
    def var(self):
        return 1

    @property
    def length_scale(self):
        return 0

    @property
    def period(self):
        return 0


class ZeroKernel(Kernel, ZeroFunction, Referentiable):
    """Constant kernel of `0`."""

    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(B.Numeric, B.Numeric)
    def __call__(self, x, y):
        if x is y:
            return Zero(B.dtype(x), B.shape(uprank(x))[0])
        else:
            return Zero(B.dtype(x),
                        B.shape(uprank(x))[0],
                        B.shape(uprank(y))[0])

    @_dispatch(B.Numeric, B.Numeric)
    @uprank
    def elwise(self, x, y):
        return B.zeros(B.dtype(x), B.shape(x)[0], 1)

    @property
    def _stationary(self):
        return True

    @property
    def var(self):
        return 0

    @property
    def length_scale(self):
        return 0

    @property
    def period(self):
        return 0


class ScaledKernel(Kernel, ScaledFunction, Referentiable):
    """Scaled kernel."""

    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(object, object)
    def __call__(self, x, y):
        return self._compute(self[0](x, y))

    @_dispatch(object, object)
    def elwise(self, x, y):
        return self._compute(self[0].elwise(x, y))

    def _compute(self, K):
        return B.multiply(self.scale, K)

    @property
    def _stationary(self):
        return self[0].stationary

    @property
    def var(self):
        return self.scale * self[0].var

    @property
    def length_scale(self):
        return self[0].length_scale

    @property
    def period(self):
        return self[0].period


class SumKernel(Kernel, SumFunction, Referentiable):
    """Sum of kernels."""

    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(object, object)
    def __call__(self, x, y):
        return B.add(self[0](x, y), self[1](x, y))

    @_dispatch(object, object)
    def elwise(self, x, y):
        return B.add(self[0].elwise(x, y), self[1].elwise(x, y))

    @property
    def _stationary(self):
        return self[0].stationary and self[1].stationary

    @property
    def var(self):
        return self[0].var + self[1].var

    @property
    def length_scale(self):
        return (self[0].var * self[0].length_scale +
                self[1].var * self[1].length_scale) / self.var

    @property
    def period(self):
        return np.inf


class ProductKernel(Kernel, ProductFunction, Referentiable):
    """Product of two kernels."""

    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(object, object)
    def __call__(self, x, y):
        return B.multiply(self[0](x, y), self[1](x, y))

    @_dispatch(object, object)
    def elwise(self, x, y):
        return B.multiply(self[0].elwise(x, y), self[1].elwise(x, y))

    @property
    def _stationary(self):
        return self[0].stationary and self[1].stationary

    @property
    def var(self):
        return self[0].var * self[1].var

    @property
    def length_scale(self):
        return B.minimum(self[0].length_scale, self[1].length_scale)

    @property
    def period(self):
        return np.inf


class StretchedKernel(Kernel, StretchedFunction, Referentiable):
    """Stretched kernel."""

    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(B.Numeric, B.Numeric)
    @uprank
    def __call__(self, x, y):
        return self[0](*self._compute(x, y))

    @_dispatch(B.Numeric, B.Numeric)
    @uprank
    def elwise(self, x, y):
        return self[0].elwise(*self._compute(x, y))

    def _compute(self, x, y):
        stretches1, stretches2 = expand(self.stretches)
        return B.divide(x, stretches1), B.divide(y, stretches2)

    @property
    def _stationary(self):
        if len(self.stretches) == 1:
            return self[0].stationary
        else:
            # NOTE: Can do something more clever here.
            return False

    @property
    def var(self):
        return self[0].var

    @property
    def length_scale(self):
        if len(self.stretches) == 1:
            return self[0].length_scale * self.stretches[0]
        else:
            # NOTE: Can do something more clever here.
            return Kernel.length_scale.fget(self)

    @property
    def period(self):
        if len(self.stretches) == 1:
            return self[0].period * self.stretches[0]
        else:
            # NOTE: Can do something more clever here.
            return Kernel.period.fget(self)

    @_dispatch(Self)
    def __eq__(self, other):
        return self[0] == other[0] and \
               tuple_equal(expand(self.stretches), expand(other.stretches))


class ShiftedKernel(Kernel, ShiftedFunction, Referentiable):
    """Shifted kernel."""

    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(B.Numeric, B.Numeric)
    @uprank
    def __call__(self, x, y):
        return self[0](*self._compute(x, y))

    @_dispatch(B.Numeric, B.Numeric)
    @uprank
    def elwise(self, x, y):
        return self[0].elwise(*self._compute(x, y))

    def _compute(self, x, y):
        shifts1, shifts2 = expand(self.shifts)
        return B.subtract(x, shifts1), B.subtract(y, shifts2)

    @property
    def _stationary(self):
        if len(self.shifts) == 1:
            return self[0].stationary
        else:
            # NOTE: Can do something more clever here.
            return False

    @property
    def var(self):
        return self[0].var

    @property
    def length_scale(self):
        if len(self.shifts) == 1:
            return self[0].length_scale
        else:
            # NOTE: Can do something more clever here.
            return Kernel.length_scale.fget(self)

    @property
    def period(self):
        return self[0].period

    @_dispatch(Self)
    def __eq__(self, other):
        return self[0] == other[0] and \
               tuple_equal(expand(self.shifts), expand(other.shifts))


class SelectedKernel(Kernel, SelectedFunction, Referentiable):
    """Kernel with particular input dimensions selected."""

    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(B.Numeric, B.Numeric)
    @uprank
    def __call__(self, x, y):
        return self[0](*self._compute(x, y))

    @_dispatch(B.Numeric, B.Numeric)
    @uprank
    def elwise(self, x, y):
        return self[0].elwise(*self._compute(x, y))

    def _compute(self, x, y):
        dims1, dims2 = expand(self.dims)
        x = x if dims1 is None else B.take(x, dims1, axis=1)
        y = y if dims2 is None else B.take(y, dims2, axis=1)
        return x, y

    @property
    def _stationary(self):
        if len(self.dims) == 1:
            return self[0].stationary
        else:
            # NOTE: Can do something more clever here.
            return False

    @property
    def var(self):
        return self[0].var

    @property
    def length_scale(self):
        length_scale = self[0].length_scale
        if B.isscalar(length_scale):
            return length_scale
        else:
            if len(self.dims) == 1:
                return B.take(length_scale, self.dims[0])
            else:
                # NOTE: Can do something more clever here.
                return Kernel.length_scale.fget(self)

    @property
    def period(self):
        period = self[0].period
        if B.isscalar(period):
            return period
        else:
            if len(self.dims) == 1:
                return B.take(period, self.dims[0])
            else:
                # NOTE: Can do something more clever here.
                return Kernel.period.fget(self)

    @_dispatch(Self)
    def __eq__(self, other):
        return self[0] == other[0] and \
               tuple_equal(expand(self.dims), expand(other.dims))


class InputTransformedKernel(Kernel, InputTransformedFunction, Referentiable):
    """Input-transformed kernel."""

    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(object, object)
    @uprank
    def __call__(self, x, y):
        return self[0](*self._compute(x, y))

    @_dispatch(object, object)
    @uprank
    def elwise(self, x, y):
        return self[0].elwise(*self._compute(x, y))

    def _compute(self, x, y):
        f1, f2 = expand(self.fs)
        x = x if f1 is None else uprank(f1(x))
        y = y if f2 is None else uprank(f2(y))
        return x, y

    @_dispatch(Self)
    def __eq__(self, other):
        return self[0] == other[0] and \
               tuple_equal(expand(self.fs), expand(other.fs))


class PeriodicKernel(Kernel, WrappedFunction, Referentiable):
    """Periodic kernel.

    Args:
        k (:class:`.kernel.Kernel`): Kernel to make periodic.
        scale (tensor): Period.
    """

    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, k, period):
        WrappedFunction.__init__(self, k)
        self._period = to_tensor(period)

    @_dispatch(B.Numeric, B.Numeric)
    @uprank
    def __call__(self, x, y):
        return self[0](*self._compute(x, y))

    @_dispatch(B.Numeric, B.Numeric)
    @uprank
    def elwise(self, x, y):
        return self[0].elwise(*self._compute(x, y))

    def _compute(self, x, y):
        def feat_map(z):
            z = B.divide(B.multiply(B.multiply(z, 2), B.pi), self.period)
            return B.concat(B.sin(z), B.cos(z), axis=1)

        return feat_map(x), feat_map(y)

    @property
    def _stationary(self):
        return self[0].stationary

    @property
    def var(self):
        return self[0].var

    @property
    def length_scale(self):
        return self[0].length_scale

    @property
    def period(self):
        return self._period

    @_dispatch(object, Formatter)
    def display(self, e, formatter):
        return '{} per {}'.format(e, formatter(self._period))

    @_dispatch(Self)
    def __eq__(self, other):
        return self[0] == other[0] and B.all(self.period == other.period)


class EQ(Kernel, Referentiable):
    """Exponentiated quadratic kernel."""

    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(B.Numeric, B.Numeric)
    @uprank
    def __call__(self, x, y):
        return Dense(self._compute(B.pw_dists2(x, y)))

    @_dispatch(B.Numeric, B.Numeric)
    @uprank
    def elwise(self, x, y):
        return self._compute(B.ew_dists2(x, y))

    def _compute(self, dists2):
        return B.exp(-0.5 * dists2)

    @property
    def _stationary(self):
        return True

    @property
    def var(self):
        return 1

    @property
    def length_scale(self):
        return 1

    @property
    def period(self):
        return np.inf

    @_dispatch(Self)
    def __eq__(self, other):
        return True


class RQ(Kernel, Referentiable):
    """Rational quadratic kernel.

    Args:
        alpha (scalar): Shape of the prior over length scales. Determines the
            weight of the tails of the kernel. Must be positive.
    """

    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, alpha):
        self.alpha = alpha

    @_dispatch(B.Numeric, B.Numeric)
    @uprank
    def __call__(self, x, y):
        return Dense(self._compute(B.pw_dists2(x, y)))

    @_dispatch(B.Numeric, B.Numeric)
    @uprank
    def elwise(self, x, y):
        return self._compute(B.ew_dists2(x, y))

    def _compute(self, dists2):
        return (1 + .5 * dists2 / self.alpha) ** (-self.alpha)

    @_dispatch(Formatter)
    def display(self, formatter):
        return 'RQ({})'.format(formatter(self.alpha))

    @property
    def _stationary(self):
        return True

    @property
    def var(self):
        return 1

    @property
    def length_scale(self):
        return 1

    @property
    def period(self):
        return np.inf

    @_dispatch(Self)
    def __eq__(self, other):
        return B.all(self.alpha == other.alpha)


class Exp(Kernel, Referentiable):
    """Exponential kernel."""

    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(B.Numeric, B.Numeric)
    @uprank
    def __call__(self, x, y):
        return Dense(B.exp(-B.pw_dists(x, y)))

    @_dispatch(B.Numeric, B.Numeric)
    @uprank
    def elwise(self, x, y):
        return B.exp(-B.ew_dists(x, y))

    @property
    def _stationary(self):
        return True

    @property
    def var(self):
        return 1

    @property
    def length_scale(self):
        return 1

    @property
    def period(self):
        return np.inf

    @_dispatch(Self)
    def __eq__(self, other):
        return True


Matern12 = Exp  #: Alias for the exponential kernel.


class Matern32(Kernel, Referentiable):
    """Matern--3/2 kernel."""

    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(B.Numeric, B.Numeric)
    @uprank
    def __call__(self, x, y):
        return Dense(self._compute(B.pw_dists(x, y)))

    @_dispatch(B.Numeric, B.Numeric)
    @uprank
    def elwise(self, x, y):
        return self._compute(B.ew_dists(x, y))

    def _compute(self, dists):
        r = 3 ** .5 * dists
        return (1 + r) * B.exp(-r)

    @property
    def _stationary(self):
        return True

    @property
    def var(self):
        return 1

    @property
    def length_scale(self):
        return 1

    @property
    def period(self):
        return np.inf

    @_dispatch(Self)
    def __eq__(self, other):
        return True


class Matern52(Kernel, Referentiable):
    """Matern--5/2 kernel."""

    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(B.Numeric, B.Numeric)
    @uprank
    def __call__(self, x, y):
        return Dense(self._compute(B.pw_dists(x, y)))

    @_dispatch(B.Numeric, B.Numeric)
    @uprank
    def elwise(self, x, y):
        return self._compute(B.ew_dists(x, y))

    def _compute(self, dists):
        r1 = 5 ** .5 * dists
        r2 = 5 * dists ** 2 / 3
        return (1 + r1 + r2) * B.exp(-r1)

    @property
    def _stationary(self):
        return True

    @property
    def var(self):
        return 1

    @property
    def length_scale(self):
        return 1

    @property
    def period(self):
        return np.inf

    @_dispatch(Self)
    def __eq__(self, other):
        return True


class Delta(Kernel, Referentiable):
    """Kronecker delta kernel.

    Args:
        epsilon (float, optional): Tolerance for equality in squared distance.
            Defaults to `1e-10`.
    """

    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, epsilon=1e-10):
        self.epsilon = epsilon

    @_dispatch(B.Numeric, B.Numeric)
    def __call__(self, x, y):
        if x is y:
            return self._eye(uprank(x))
        else:
            return Dense(self._compute(B.pw_dists2(uprank(x), uprank(y))))

    @_dispatch(Unique, Unique)
    def __call__(self, x, y):
        x, y = x.get(), y.get()
        if x is y:
            return self._eye(uprank(x))
        else:
            x, y = uprank(x), uprank(y)
            return Zero(B.dtype(x), B.shape(x)[0], B.shape(y)[0])

    @_dispatch(Unique, object)
    def __call__(self, x, y):
        x, y = uprank(x.get()), uprank(y)
        return Zero(B.dtype(x), B.shape(x)[0], B.shape(y)[0])

    @_dispatch(object, Unique)
    def __call__(self, x, y):
        x, y = uprank(x), uprank(y.get())
        return Zero(B.dtype(x), B.shape(x)[0], B.shape(y)[0])

    @_dispatch(Unique, Unique)
    def elwise(self, x, y):
        x, y = x.get(), y.get()
        if x is y:
            return One(B.dtype(x), B.shape(uprank(x))[0], 1)
        else:
            return Zero(B.dtype(x), B.shape(uprank(x))[0], 1)

    @_dispatch(Unique, object)
    def elwise(self, x, y):
        x = x.get()
        return Zero(B.dtype(x), B.shape(uprank(x))[0], 1)

    @_dispatch(object, Unique)
    def elwise(self, x, y):
        return Zero(B.dtype(x), B.shape(uprank(x))[0], 1)

    @_dispatch(B.Numeric, B.Numeric)
    def elwise(self, x, y):
        if x is y:
            return One(B.dtype(x), B.shape(uprank(x))[0], 1)
        else:
            return self._compute(B.ew_dists2(uprank(x), uprank(y)))

    def _eye(self, x):
        return UniformlyDiagonal(B.cast(B.dtype(x), 1), B.shape(x)[0])

    def _compute(self, dists2):
        dtype = B.dtype(dists2)
        return B.cast(dtype, B.lt(dists2, B.cast(dtype, self.epsilon)))

    @property
    def _stationary(self):
        return True

    @property
    def var(self):
        return 1

    @property
    def length_scale(self):
        return 0

    @property
    def period(self):
        return np.inf

    @_dispatch(Self)
    def __eq__(self, other):
        return self.epsilon == other.epsilon


class FixedDelta(Kernel, Referentiable):
    """Kronecker delta kernel that produces a diagonal matrix with given
    noises if and only if the inputs are identical and of the right shape.

    Args:
        noises (vector): Noises.
    """

    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, noises):
        self.noises = noises

    @_dispatch(B.Numeric, B.Numeric)
    def __call__(self, x, y):
        if x is y and B.shape(uprank(x))[0] == B.shape(self.noises)[0]:
            return Diagonal(self.noises)
        else:
            x, y = uprank(x), uprank(y)
            return Zero(B.dtype(x), B.shape(x)[0], B.shape(y)[0])

    @_dispatch(B.Numeric, B.Numeric)
    def elwise(self, x, y):
        if x is y and B.shape(B.uprank(x))[0] == B.shape(self.noises)[0]:
            return B.uprank(self.noises)
        else:
            x = B.uprank(x)
            return Zero(B.dtype(x), B.shape(x)[0], 1)

    @property
    def _stationary(self):
        return True

    @property
    def var(self):
        return 1

    @property
    def length_scale(self):
        return 0

    @property
    def period(self):
        return np.inf

    @_dispatch(Self)
    def __eq__(self, other):
        return B.all(self.noises == other.noises)


class Linear(Kernel, Referentiable):
    """Linear kernel."""

    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(B.Numeric, B.Numeric)
    def __call__(self, x, y):
        if x is y:
            return LowRank(uprank(x))
        else:
            return LowRank(left=uprank(x),
                           right=uprank(y))

    @_dispatch(B.Numeric, B.Numeric)
    @uprank
    def elwise(self, x, y):
        return B.expand_dims(B.sum(B.multiply(x, y), axis=1), axis=1)

    @property
    def _stationary(self):
        return False

    @property
    def period(self):
        return np.inf

    @_dispatch(Self)
    def __eq__(self, other):
        return True


class DecayingKernel(Kernel, Referentiable):
    """Decaying kernel.

    Args:
        alpha (tensor): Shape of the gamma distribution governing the
            distribution of decaying exponentials.
        beta (tensor): Rate of the gamma distribution governing the
            distribution of decaying exponentials.
    """

    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, alpha, beta):
        self.alpha = alpha
        self.beta = beta

    @_dispatch(B.Numeric, B.Numeric)
    @uprank
    def __call__(self, x, y):
        return B.divide(self._compute_beta_raised(),
                        B.power(B.pw_sums(B.add(x, self.beta), y), self.alpha))

    @_dispatch(B.Numeric, B.Numeric)
    @uprank
    def elwise(self, x, y):
        return B.divide(self._compute_beta_raised(),
                        B.power(B.ew_sums(B.add(x, self.beta), y), self.alpha))

    def _compute_beta_raised(self):
        beta_norm = B.sqrt(B.maximum(B.sum(B.power(self.beta, 2)),
                                     B.cast(B.dtype(self.beta), 1e-30)))
        return B.power(beta_norm, self.alpha)

    @_dispatch(Formatter)
    def display(self, formatter):
        return 'DecayingKernel({}, {})'.format(formatter(self.alpha),
                                               formatter(self.beta))

    @property
    def period(self):
        return np.inf

    @_dispatch(Self)
    def __eq__(self, other):
        return B.all(self.alpha == other.alpha) and \
               B.all(self.beta == other.beta)


class LogKernel(Kernel, Referentiable):
    """Logarithm kernel."""

    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(B.Numeric, B.Numeric)
    @uprank
    def __call__(self, x, y):
        dists = B.maximum(B.pw_dists(x, y), 1e-10)
        return B.divide(B.log(dists + 1), dists)

    @_dispatch(B.Numeric, B.Numeric)
    @uprank
    def elwise(self, x, y):
        dists = B.maximum(B.ew_dists(x, y), 1e-10)
        return B.divide(B.log(dists + 1), dists)

    @_dispatch(Formatter)
    def display(self, formatter):
        return 'LogKernel()'

    @property
    def _stationary(self):
        return True

    @property
    def var(self):
        return 1

    @property
    def length_scale(self):
        return np.inf

    @property
    def period(self):
        return np.inf

    @_dispatch(Self)
    def __eq__(self, other):
        return True


class PosteriorKernel(Kernel, Referentiable):
    """Posterior kernel.

    Args:
        k_ij (:class:`.kernel.Kernel`): Kernel between processes
            corresponding to the left input and the right input respectively.
        k_zi (:class:`.kernel.Kernel`): Kernel between processes
            corresponding to the data and the left input respectively.
        k_zj (:class:`.kernel.Kernel`): Kernel between processes
            corresponding to the data and the right input respectively.
        z (input): Locations of data.
        K_z (:class:`.matrix.Dense`): Kernel matrix of data.
    """

    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, k_ij, k_zi, k_zj, z, K_z):
        self.k_ij = k_ij
        self.k_zi = k_zi
        self.k_zj = k_zj
        self.z = z
        self.K_z = matrix(K_z)

    @_dispatch(object, object)
    def __call__(self, x, y):
        return B.schur(self.k_ij(x, y),
                       self.k_zi(self.z, x),
                       self.K_z,
                       self.k_zj(self.z, y))

    @_dispatch(object, object)
    def elwise(self, x, y):
        qf_diag = B.qf_diag(self.K_z,
                            self.k_zi(self.z, x),
                            self.k_zj(self.z, y))
        return B.subtract(self.k_ij.elwise(x, y),
                          B.expand_dims(qf_diag, axis=1))


class CorrectiveKernel(Kernel, Referentiable):
    """Kernel that adds the corrective variance in sparse conditioning.

    Args:
        k_zi (:class:`.kernel.Kernel`): Kernel between the processes
            corresponding to the left input and the inducing points
            respectively.
        k_zj (:class:`.kernel.Kernel`): Kernel between the processes
            corresponding to the right input and the inducing points
            respectively.
        z (input): Locations of the inducing points.
        A (tensor): Corrective matrix.
        L (tensor): Kernel matrix of the inducing points.
    """
    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, k_zi, k_zj, z, A, K_z):
        self.k_zi = k_zi
        self.k_zj = k_zj
        self.z = z
        self.A = A
        self.L = B.cholesky(matrix(K_z))

    @_dispatch(object, object)
    def __call__(self, x, y):
        return B.qf(self.A,
                    B.trisolve(self.L, self.k_zi(self.z, x)),
                    B.trisolve(self.L, self.k_zj(self.z, y)))

    @_dispatch(object, object)
    def elwise(self, x, y):
        return B.qf_diag(self.A,
                         B.trisolve(self.L, self.k_zi(self.z, x)),
                         B.trisolve(self.L, self.k_zj(self.z, y)))[:, None]


def dkx(k_elwise, i):
    """Construct the derivative of a kernel with respect to its first
    argument.

    Args:
        k_elwise (function): Function that performs element-wise computation
            of the kernel.
        i (int): Dimension with respect to which to compute the derivative.

    Returns:
        function: Derivative of the kernel with respect to its first argument.
    """

    def _dkx(x, y):
        with tf.GradientTape() as t:
            # Get the numbers of inputs.
            nx = B.shape(x)[0]
            ny = B.shape(y)[0]

            # Copy the input `ny` times to efficiently compute many derivatives.
            xis = tf.identity_n([x[:, i:i + 1]] * ny)
            t.watch(xis)

            # Tile inputs for batched computation.
            x = B.tile(x, ny, 1)
            y = B.reshape(B.tile(y, 1, nx), ny * nx, -1)

            # Insert tracked dimension, which is different for every tile.
            xi = B.concat(*xis, axis=0)
            x = B.concat(x[:, :i], xi, x[:, i + 1:], axis=1)

            # Perform the derivative computation.
            out = dense(k_elwise(x, y))
            grads = t.gradient(out, xis, unconnected_gradients='zero')
            return B.concat(*grads, axis=1)

    return _dkx


def dkx_elwise(k_elwise, i):
    """Construct the element-wise derivative of a kernel with respect to
    its first argument.

    Args:
        k_elwise (function): Function that performs element-wise computation
            of the kernel.
        i (int): Dimension with respect to which to compute the derivative.

    Returns:
        function: Element-wise derivative of the kernel with respect to its
            first argument.
    """

    def _dkx_elwise(x, y):
        with tf.GradientTape() as t:
            xi = x[:, i:i + 1]
            t.watch(xi)
            x = B.concat(x[:, :i], xi, x[:, i + 1:], axis=1)
            out = dense(k_elwise(x, y))
            return t.gradient(out, xi, unconnected_gradients='zero')

    return _dkx_elwise


def dky(k_elwise, i):
    """Construct the derivative of a kernel with respect to its second
    argument.

    Args:
        k_elwise (function): Function that performs element-wise computation
            of the kernel.
        i (int): Dimension with respect to which to compute the derivative.

    Returns:
        function: Derivative of the kernel with respect to its second argument.
    """

    def _dky(x, y):
        with tf.GradientTape() as t:
            # Get the numbers of inputs.
            nx = B.shape(x)[0]
            ny = B.shape(y)[0]

            # Copy the input `nx` times to efficiently compute many derivatives.
            yis = tf.identity_n([y[:, i:i + 1]] * nx)
            t.watch(yis)

            # Tile inputs for batched computation.
            x = B.reshape(B.tile(x, 1, ny), nx * ny, -1)
            y = B.tile(y, nx, 1)

            # Insert tracked dimension, which is different for every tile.
            yi = B.concat(*yis, axis=0)
            y = B.concat(y[:, :i], yi, y[:, i + 1:], axis=1)

            # Perform the derivative computation.
            out = dense(k_elwise(x, y))
            grads = t.gradient(out, yis, unconnected_gradients='zero')
            return B.transpose(B.concat(*grads, axis=1))

    return _dky


def dky_elwise(k_elwise, i):
    """Construct the element-wise derivative of a kernel with respect to
    its second argument.

    Args:
        k_elwise (function): Function that performs element-wise computation
            of the kernel.
        i (int): Dimension with respect to which to compute the derivative.

    Returns:
        function: Element-wise derivative of the kernel with respect to its
            second argument.
    """

    def _dky_elwise(x, y):
        with tf.GradientTape() as t:
            yi = y[:, i:i + 1]
            t.watch(yi)
            y = B.concat(y[:, :i], yi, y[:, i + 1:], axis=1)
            out = dense(k_elwise(x, y))
            return t.gradient(out, yi, unconnected_gradients='zero')

    return _dky_elwise


def perturb(x):
    """Slightly perturb a tensor.

    Args:
        x (tensor): Tensor to perturb.

    Returns:
        tensor: `x`, but perturbed.
    """
    dtype = convert(B.dtype(x), B.NPDType)
    if dtype == np.float64:
        return 1e-20 + x * (1 + 1e-14)
    elif dtype == np.float32:
        return 1e-20 + x * (1 + 1e-7)
    else:
        raise ValueError('Cannot perturb a tensor of data type {}.'
                         ''.format(B.dtype(x)))


class DerivativeKernel(Kernel, DerivativeFunction, Referentiable):
    """Derivative of kernel."""
    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(B.Numeric, B.Numeric)
    @uprank
    def __call__(self, x, y):
        i, j = expand(self.derivs)
        k = self[0]

        # Prevent that `x` equals `y` to stabilise nested gradients.
        y = perturb(y)

        if i is not None and j is not None:
            # Derivative with respect to both `x` and `y`.
            return Dense(dky(dkx_elwise(k.elwise, i), j)(x, y))

        elif i is not None and j is None:
            # Derivative with respect to `x`.
            return Dense(dkx(k.elwise, i)(x, y))

        elif i is None and j is not None:
            # Derivative with respect to `y`.
            return Dense(dky(k.elwise, j)(x, y))

        else:
            raise RuntimeError('No derivative specified.')

    @_dispatch(B.Numeric, B.Numeric)
    @uprank
    def elwise(self, x, y):
        i, j = expand(self.derivs)
        k = self[0]

        # Prevent that `x` equals `y` to stabilise nested gradients.
        y = perturb(y)

        if i is not None and j is not None:
            # Derivative with respect to both `x` and `y`.
            return dky_elwise(dkx_elwise(k.elwise, i), j)(x, y)

        elif i is not None and j is None:
            # Derivative with respect to `x`.
            return dkx_elwise(k.elwise, i)(x, y)

        elif i is None and j is not None:
            # Derivative with respect to `y`.
            return dky_elwise(k.elwise, j)(x, y)

        else:
            raise RuntimeError('No derivative specified.')

    @property
    def _stationary(self):
        # NOTE: In the one-dimensional case, if derivatives with respect to both
        #     arguments are taken, then the result is in fact stationary.
        return False

    @_dispatch(Self)
    def __eq__(self, other):
        return self[0] == other[0] and \
               tuple_equal(expand(self.derivs), expand(other.derivs))


class TensorProductKernel(Kernel, TensorProductFunction, Referentiable):
    """Tensor product kernel."""
    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(B.Numeric, B.Numeric)
    def __call__(self, x, y):
        f1, f2 = expand(self.fs)
        if x is y and f1 is f2:
            return LowRank(uprank(f1(uprank(x))))
        else:
            return LowRank(left=uprank(f1(uprank(x))),
                           right=uprank(f2(uprank(y))))

    @_dispatch(B.Numeric, B.Numeric)
    @uprank
    def elwise(self, x, y):
        f1, f2 = expand(self.fs)
        return B.multiply(uprank(f1(x)), uprank(f2(y)))

    @_dispatch(Self)
    def __eq__(self, other):
        return tuple_equal(expand(self.fs), expand(other.fs))


class ReversedKernel(Kernel, WrappedFunction, Referentiable):
    """Reversed kernel.

    Evaluates with its arguments reversed.
    """
    _dispatch = Dispatcher(in_class=Self)

    @_dispatch(object, object)
    def __call__(self, x, y):
        return B.transpose(self[0](y, x))

    @_dispatch(object, object)
    def elwise(self, x, y):
        return self[0].elwise(y, x)

    @property
    def _stationary(self):
        return self[0].stationary

    @property
    def var(self):
        return self[0].var

    @property
    def length_scale(self):
        return self[0].length_scale

    @property
    def period(self):
        return self[0].period

    @_dispatch(object, Formatter)
    def display(self, e, formatter):
        return 'Reversed({})'.format(e)

    @_dispatch(Self)
    def __eq__(self, other):
        return self[0] == other[0]


@need_parens.extend_multi((Function, ReversedKernel),
                          ({WrappedFunction, JoinFunction}, ReversedKernel))
def need_parens(el, parent): return False


@need_parens.extend(ReversedKernel, ProductFunction)
def need_parens(el, parent): return False


# Periodicise kernels.

@_dispatch(Kernel, object)
def periodicise(a, b): return PeriodicKernel(a, b)


@_dispatch(ZeroKernel, object)
def periodicise(a, b): return a


# Reverse kernels.

@_dispatch(Kernel)
def reverse(a): return a if a.stationary else ReversedKernel(a)


@_dispatch(ReversedKernel)
def reverse(a): return a[0]


@_dispatch.multi((ZeroKernel,), (OneKernel,))
def reverse(a): return a


@_dispatch(ShiftedKernel)
def reverse(a): return shift(reversed(a[0]), *reversed(a.shifts))


@_dispatch(StretchedKernel)
def reverse(a): return stretch(reversed(a[0]), *reversed(a.stretches))


@_dispatch(InputTransformedKernel)
def reverse(a): return transform(reversed(a[0]), *reversed(a.fs))


@_dispatch(SelectedKernel)
def reverse(a): return select(reversed(a[0]), *reversed(a.dims))


# Propagate reversal.

@_dispatch(SumKernel)
def reverse(a): return add(reverse(a[0]), reverse(a[1]))


@_dispatch(ProductKernel)
def reverse(a): return mul(reverse(a[0]), reverse(a[1]))


@_dispatch(ScaledKernel)
def reverse(a): return mul(a.scale, reversed(a[0]))


# Make shifting synergise with reversal.

@shift.extend(Kernel, [object])
def shift(a, *shifts):
    if a.stationary and len(shifts) == 1:
        return a
    else:
        return ShiftedKernel(a, *shifts)


@shift.extend(ZeroKernel, [object])
def shift(a, *shifts): return a


@shift.extend(ShiftedKernel, [object])
def shift(a, *shifts):
    return shift(a[0], *broadcast(operator.add, a.shifts, shifts))



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function

import logging

from plum import Dispatcher, Referentiable, Self

__all__ = []

log = logging.getLogger(__name__)


def replace_at(tup, i, val):
    """Replace a value in a tuple.

    Args:
        tup (tuple): Tuple to replace value in.
        i (int): Index to replace value at.
        val (obj): New value.

    Returns:
        tuple: Tuple with the value replaced.
    """
    listed = list(tup)
    listed[i] = val
    return tuple(listed)


class Rule(object):
    """A rule for an :class:`.lazy.LazyTensor`.

    Note:
        For performance reasons, `indices` must already be resolved!

    Args:
        pattern (tuple): Rule pattern. Each element must be a `None` or an
            element from `indices`. A `None` represents a wildcard.
        indices (set): Values to match wildcards with. These must already have
            been resolved.
        builder (function): Function that builds the element. Only the wildcard
            elements are fed to the function.
    """

    def __init__(self, pattern, indices, builder):
        self.pattern = pattern
        self.indices = set(indices)
        self.builder = builder
        self._args = None

    def applies(self, key):
        """Check whether this rule applies to a certain key.

        Args:
            key (tuple): Key.

        Returns:
            bool: Boolean indicating whether the rule applies.
        """
        # Already keep track of arguments for building.
        self._args = ()

        for (i, j) in zip(self.pattern, key):
            if i is None:
                if j not in self.indices:
                    return False
                else:
                    self._args += (j,)
            elif i != j:
                return False
        return True

    def build(self):
        """After checking, build the element."""
        return self.builder(*self._args)

    def __repr__(self):
        return 'Rule(pattern={!r}, indices={!r}, builder={!r})' \
               ''.format(self.pattern, self.indices, self.builder)


class LazyTensor(Referentiable):
    """A lazy tensor that indexes by the identity of its indices.

    Args:
        d (int): Rank of the tensor.
    """
    _dispatch = Dispatcher(in_class=Self)

    def __init__(self, rank):
        self.rank = rank
        self._rules = {}
        self._store = {}

    @_dispatch(object)
    def _resolve_index(self, i):
        return id(i)

    @_dispatch(type(None))
    def _resolve_index(self, i):
        return None

    @_dispatch(int)
    def _resolve_index(self, i):
        return i

    @_dispatch(object)
    def _resolve_key(self, key):
        return (self._resolve_index(key),) * self.rank

    @_dispatch({tuple, reversed})
    def _resolve_key(self, key):
        return tuple(self._resolve_index(i) for i in key)

    def __setitem__(self, key, value):
        return self._set(self._resolve_key(key), value)

    def __getitem__(self, key):
        return self._get(self._resolve_key(key))

    def _set(self, key, value):
        self._store[key] = value

    def _get(self, key):
        try:
            return self._store[key]
        except KeyError:
            pass

        # Finally, try building the element.
        value = self._build(key)
        self._store[key] = value
        return value

    def _build(self, key):
        # NOTE: The building patterns should go from least specific to most
        # specific.
        # Try a universal match.
        building_patterns = [(None,) * self.rank]
        # Try single dimension patterns.
        building_patterns += [replace_at(key, i, None) for i in
                              range(self.rank)]
        # Try key.
        building_patterns += [key]

        for pattern in building_patterns:
            # Check if a rules exists for the pattern.
            if pattern in self._rules:
                # Rules exist! Build with the first one that applies.
                for rule in self._rules[pattern]:
                    if rule.applies(key):
                        return rule.build()

        # Unable to build value for key.
        raise RuntimeError('Could not build value for key "{}".'.format(key))

    @_dispatch([object])
    def add_rule(self, pattern, indices, builder):
        """Add a building rule.

        See :class:`.lazy.Rule`.

        Note:
            For performance reasons, `indices` must already be resolved!

        Args:
            pattern (tuple): Rule pattern. Each element must be a `None` or an
                element from `indices`. A `None` represents a wildcard.
            indices (set): Values to match wildcards with. These must already have
                been resolved.
            builder (function): Function that builds the element. Only the wildcard
                elements are fed to the function.
        """
        pattern = self._resolve_key(pattern)
        # IMPORTANT: This assumes that the indices already have been resolved!
        rule = Rule(pattern, indices, builder)
        try:
            self._rules[pattern].append(rule)
        except KeyError:
            self._rules[pattern] = [rule]


class LazyVector(LazyTensor):
    """A lazy vector."""

    def __init__(self):
        LazyTensor.__init__(self, 1)


class LazyMatrix(LazyTensor):
    """A lazy matrix."""

    def __init__(self):
        LazyTensor.__init__(self, 2)
